#
#   Copyright 2024 Hopsworks AB
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
from __future__ import annotations

import re
from datetime import date, datetime
from typing import Any, Dict, List, Optional, Union

import pandas as pd
from hopsworks_common.client.exceptions import FeatureStoreException
from hsfs import feature_group, feature_view
from hsfs.core import feature_monitoring_config as fmc
from hsfs.core import feature_monitoring_config_api, monitoring_window_config_engine
from hsfs.core import monitoring_window_config as mwc
from hsfs.core.feature_monitoring_result import FeatureMonitoringResult
from hsfs.core.feature_monitoring_result_engine import FeatureMonitoringResultEngine
from hsfs.core.job import Job
from hsfs.core.job_api import JobApi


class FeatureMonitoringConfigEngine:
    def __init__(
        self,
        feature_store_id: int,
        feature_group_id: Optional[int] = None,
        feature_view_name: Optional[str] = None,
        feature_view_version: Optional[int] = None,
        **kwargs,
    ):
        """Business logic for feature monitoring configuration.

        This class encapsulates the business logic for feature monitoring configuration.
        It is responsible for routing methods from the public python API to the
        appropriate REST calls. It should also contain validation and error handling logic
        for payloads or default object. Additionally, it contains logic necessary
        to run the feature monitoring job, including taking a monitoring window configuration
        and fetching the associated data.

        Attributes:
            feature_store_id: int. Id of the respective Feature Store.
            feature_group_id: int. Id of the feature group, if monitoring a feature group.
            feature_view_name: str. Name of the feature view, if monitoring a feature view.
            feature_view_version: int. Version of the feature view, if monitoring a feature view.
        """
        self._feature_store_id = feature_store_id
        self._feature_group_id = feature_group_id
        self._feature_view_name = feature_view_name
        self._feature_view_version = feature_view_version

        self._feature_monitoring_config_api = (
            feature_monitoring_config_api.FeatureMonitoringConfigApi(
                feature_store_id=feature_store_id,
                feature_group_id=feature_group_id,
                feature_view_name=feature_view_name,
                feature_view_version=feature_view_version,
            )
        )
        self._job_api = JobApi()
        self._monitoring_window_config_engine = (
            monitoring_window_config_engine.MonitoringWindowConfigEngine()
        )
        self._result_engine = FeatureMonitoringResultEngine(
            feature_store_id=feature_store_id,
            feature_group_id=feature_group_id,
            feature_view_name=feature_view_name,
            feature_view_version=feature_view_version,
        )

        # Should we promote the variables below to static?
        self._VALID_CATEGORICAL_METRICS = [
            "completeness",
            "num_records_non_null",
            "num_records_null",
            "distinctness",
            "entropy",
            "uniqueness",
            "approximate_num_distinct_values",
            "exact_num_distinct_values",
        ]
        self._VALID_FRACTIONAL_METRICS = [
            "completeness",
            "num_records_non_null",
            "num_records_null",
            "distinctness",
            "entropy",
            "uniqueness",
            "approximate_num_distinct_values",
            "exact_num_distinct_values",
            "mean",
            "max",
            "min",
            "sum",
            "std_dev",
            "count",
        ]

    def validate_statistics_comparison_config(
        self,
        metric: str,
        threshold: Union[float, int],
        id: Optional[int] = None,
        relative: bool = False,
        strict: bool = False,
    ) -> Dict[str, Any]:
        """Validates the statistics comparison config.

        Args:
            metric: str, required
                Statistical metric to perform comparison on.
            threshold: float, required
                If statistics difference is above threshold, trigger an alert if configured.
            relative: bool, optional
                If true, compute the relative difference between the detection and reference statistics.
                Defaults to false.
            strict: bool, optional
                If true, the statistics difference must be strictly above threshold to trigger an alert.
                Defaults to false.

        Raises:
            ValueError: If the statistics comparison config is invalid.
        """
        if not isinstance(relative, bool):
            raise ValueError("relative must be a boolean value.")

        if not isinstance(strict, bool):
            raise ValueError("strict must be a boolean value.")

        if not isinstance(threshold, (int, float)):
            raise TypeError("threshold must be a numeric value.")

        if not isinstance(metric, str):
            raise TypeError(
                "metric must be a string value. "
                "Check the documentation for a list of supported metrics."
            )
        # TODO: [FSTORE-1205] Add more validation logic based on detection and reference window config.
        if (
            metric.lower() not in self._VALID_CATEGORICAL_METRICS
            and metric.lower() not in self._VALID_FRACTIONAL_METRICS
        ):
            raise ValueError(
                f"Invalid metric {metric.lower()}. " "Supported metrics are {}.".format(
                    set(self._VALID_FRACTIONAL_METRICS).union(
                        set(self._VALID_CATEGORICAL_METRICS)
                    )
                )
            )

        return {
            "id": id,
            "metric": metric.upper(),
            "threshold": threshold,
            "relative": relative,
            "strict": strict,
        }

    def validate_config_name(self, name: str):
        if not isinstance(name, str):
            raise TypeError("Invalid config name. Config name must be a string.")
        if len(name) > 64:
            raise ValueError(
                "Invalid config name. Config name must be less than 64 characters."
            )
        if not re.match(r"^[\w]*$", name):
            raise ValueError(
                "Invalid config name. Config name must be alphanumeric or underscore."
            )

    def validate_description(self, description: Optional[str]):
        if description is not None and not isinstance(description, str):
            raise TypeError("Invalid description. Description must be a string.")
        if description is not None and len(description) > 256:
            raise ValueError(
                "Invalid description. Description must be less than 256 characters."
            )

    def validate_feature_name(
        self, feature_name: Optional[str], valid_feature_names: List[str]
    ):
        if feature_name is not None and not isinstance(feature_name, str):
            raise TypeError("Invalid feature name. Feature name must be a string.")
        if feature_name is not None and feature_name not in valid_feature_names:
            raise ValueError(
                f"Invalid feature name. Feature name must be one of {valid_feature_names}."
            )

    def save(
        self, config: "fmc.FeatureMonitoringConfig"
    ) -> "fmc.FeatureMonitoringConfig":
        """Saves a feature monitoring config.

        Args:
            config: FeatureMonitoringConfig, required
                The feature monitoring config to save.

        Returns:
            FeatureMonitoringConfig The saved feature monitoring config.

        Raises:
            FeatureStoreException: If the config is already registered.
        """
        if config._id is not None:
            raise FeatureStoreException(
                "Cannot save a config that is already registered."
                " Please use update() instead."
            )
        return self._feature_monitoring_config_api.create(config)

    def update(
        self, config: "fmc.FeatureMonitoringConfig"
    ) -> "fmc.FeatureMonitoringConfig":
        """Updates a feature monitoring config.

        Args:
            config: FeatureMonitoringConfig, required
                The feature monitoring config to update.

        Returns:
            FeatureMonitoringConfig The updated feature monitoring config.

        Raises:
            FeatureStoreException: If the config is not registered.
        """
        if config._id is None:
            raise FeatureStoreException(
                "Cannot update a config that is not registered."
                " Please use save() instead."
            )
        return self._feature_monitoring_config_api.update(config)

    def delete(self, config_id: int) -> None:
        """Deletes a feature monitoring config.

        Args:
            config_id: int, required
                The id of the feature monitoring config to delete.
        """
        self._feature_monitoring_config_api.delete(config_id=config_id)

    def get_feature_monitoring_configs(
        self,
        name: Optional[str] = None,
        feature_name: Optional[str] = None,
        config_id: Optional[int] = None,
    ) -> Union[
        "fmc.FeatureMonitoringConfig", List["fmc.FeatureMonitoringConfig"], None
    ]:
        """Fetch feature monitoring configuration by entity, name or feature name.

        If no arguments are provided, fetches all feature monitoring configurations
        attached to the given entity. If a name is provided, it fetches a single configuration
        and returns None if not found. If a feature name is provided, it fetches all
        configurations attached to that feature (not including those attached to the full
        entity) and returns an empty list if none are found. If a config_id is provided,
        it fetches a single configuration and returns None if not found.

        Args:
            name: str, optional
                If provided, fetch only configuration with given name.
                Defaults to None.
            feature_name: str, optional
                If provided, fetch all configurations attached to a specific feature.
                Defaults to None.
            config_id: int, optional
                If provided, fetch only configuration with given id.
                Defaults to None.

        Raises:
            ValueError: If both name and feature_name are provided.
            TypeError: If name or feature_name are not strings.

        Returns:
            FeatureMonitoringConfig or List[FeatureMonitoringConfig] The monitoring
            configuration(s).
        """
        if any(
            [
                name and feature_name,
                feature_name and config_id,
                config_id and name,
            ]
        ):
            raise ValueError("Provide at most one of name, feature_name, or config_id.")

        if name is not None:
            if not isinstance(name, str):
                raise TypeError("name must be a string or None.")
            return self._feature_monitoring_config_api.get_by_name(name=name)
        elif feature_name is not None:
            if not isinstance(feature_name, str):
                raise TypeError("feature_name must be a string or None.")
            return self._feature_monitoring_config_api.get_by_feature_name(
                feature_name=feature_name
            )
        elif config_id is not None:
            if not isinstance(config_id, int):
                raise TypeError("config_id must be an integer or None.")
            return self._feature_monitoring_config_api.get_by_id(config_id=config_id)

        return self._feature_monitoring_config_api.get_by_entity()

    def trigger_monitoring_job(
        self,
        job_name: str,
    ) -> Job:
        """Make a REST call to start an execution of the monitoring job.

        Args:
            job_name: Name of the job to trigger.

        Returns:
            Job object.
        """
        self._job_api.launch(name=job_name)

        return self._job_api.get(name=job_name)

    def get_monitoring_job(
        self,
        job_name: str,
    ) -> Job:
        """Make a REST call to fetch the job entity.

        Args:
            job_name: Name of the job to trigger.

        Returns:
            `Job` A Hopsworks job with its metadata and execution history.
        """
        return self._job_api.get(name=job_name)

    def run_feature_monitoring(
        self,
        entity: Union[feature_group.FeatureGroup, "feature_view.FeatureView"],
        config_name: str,
    ) -> List[FeatureMonitoringResult]:
        """Main function used by the job to actually perform the monitoring.

        Args:
            entity: Union[feature_group.FeatureGroup, "feature_view.FeatureView"]
                Featuregroup or Featureview object containing the feature to monitor.
            config_name: str: name of the monitoring config.

        Returns:
            List[FeatureMonitoringResult]: A list of result object describing the
                outcome of the monitoring.
        """
        config = self._feature_monitoring_config_api.get_by_name(config_name)

        assert config is not None, "Feature monitoring config not found."

        # TODO: [FSTORE-1206] Parallelize both single_window_monitoring calls and wait
        detection_statistics = (
            self._monitoring_window_config_engine.run_single_window_monitoring(
                entity=entity,
                monitoring_window_config=config.detection_window_config,
                feature_name=config.feature_name,
            )
        )

        specific_value, reference_statistics = None, None
        if config.reference_window_config is not None:
            if (
                config.reference_window_config.window_config_type
                == mwc.WindowConfigType.SPECIFIC_VALUE
            ):
                specific_value = config.reference_window_config.specific_value
            else:
                reference_statistics = (
                    self._monitoring_window_config_engine.run_single_window_monitoring(
                        entity=entity,
                        monitoring_window_config=config.reference_window_config,
                        feature_name=config.feature_name,
                    )
                )

        return self._result_engine.run_and_save_statistics_comparison(
            fm_config=config,
            detection_statistics=detection_statistics,
            reference_statistics=reference_statistics,
            specific_value=specific_value,
        )

    def _build_default_statistics_monitoring_config(
        self,
        name: str,
        feature_name: Optional[str] = None,
        start_date_time: Optional[Union[str, int, date, datetime, pd.Timestamp]] = None,
        description: Optional[str] = None,
        valid_feature_names: Optional[List[str]] = None,
        end_date_time: Optional[Union[str, int, date, datetime, pd.Timestamp]] = None,
        cron_expression: Optional[str] = "0 0 12 ? * * *",
    ) -> "fmc.FeatureMonitoringConfig":
        """Builds the default scheduled statistics config, default detection window is full snapshot.

        Args:
            name: str, required
                Name of the feature monitoring configuration, must be unique for
                the feature view or feature group.
            feature_name: str, optional
                If provided, compute statistics only for this feature. If none,
                defaults, compute statistics for all features.
            cron_expression: str, optional
                cron expression defining the schedule for computing statistics. The expression
                must be in UTC timezone and based on Quartz cron syntax. Default is '0 0 12 ? * * *',
                every day at 12pm UTC.
            start_date_time: Union[str, int, date, datetime, pd.Timestamp], optional
                Statistics will start being computed on schedule from that time.
            end_date_time: Union[str, int, date, datetime, pd.Timestamp], optional
                Statistics will stop being computed on schedule from that time.
            description: str, optional
                Description of the feature monitoring configuration.
            valid_feature_names: List[str], optional
                List of the feature names for the feature view or feature group.

        Returns:
            FeatureMonitoringConfig A Feature Monitoring Configuration to compute
              the statistics of a snapshot of all data present in the entity.
        """
        self.validate_config_name(name)
        self.validate_description(description)
        if feature_name is not None and valid_feature_names is not None:
            self.validate_feature_name(feature_name, valid_feature_names)

        return fmc.FeatureMonitoringConfig(
            feature_store_id=self._feature_store_id,
            feature_group_id=self._feature_group_id,
            feature_view_name=self._feature_view_name,
            feature_view_version=self._feature_view_version,
            name=name,
            description=description,
            feature_name=feature_name,
            feature_monitoring_type=fmc.FeatureMonitoringType.STATISTICS_COMPUTATION,
            job_schedule={
                "start_date_time": start_date_time or datetime.now(),
                "end_date_time": end_date_time,
                "cron_expression": cron_expression,
                "enabled": True,
            },
        ).with_detection_window()

    def _build_default_feature_monitoring_config(
        self,
        name: str,
        feature_name: str,
        start_date_time: Optional[Union[str, int, date, datetime, pd.Timestamp]] = None,
        description: Optional[str] = None,
        valid_feature_names: Optional[List[str]] = None,
        end_date_time: Optional[Union[str, int, date, datetime, pd.Timestamp]] = None,
        cron_expression: Optional[str] = "0 0 12 ? * * *",
    ) -> "fmc.FeatureMonitoringConfig":
        """Builds the default scheduled statistics config, default detection window is full snapshot.

        Args:
            name: str, required
                Name of the feature monitoring configuration, must be unique for
                the feature view or feature group.
            feature_name: str, optional
                If provided, compute statistics only for this feature. If none,
                defaults, compute statistics for all features.
            cron_expression: str, optional
                cron expression defining the schedule for computing statistics. The expression
                must be in UTC timezone and based on Quartz cron syntax. Default is '0 0 12 ? * * *',
                every day at 12pm UTC.
            start_date_time: Union[str, int, date, datetime, pd.Timestamp], optional
                Statistics will start being computed on schedule from that time.
            end_date_time: Union[str, int, date, datetime, pd.Timestamp], optional
                Statistics will stop being computed on schedule from that time.
            description: str, optional
                Description of the feature monitoring configuration.
            valid_feature_names: List[str], optional
                List of the feature names for the feature view or feature group.

        Returns:
            FeatureMonitoringConfig A Feature Monitoring Configuration to compute
              the statistics of a snapshot of all data present in the entity.
        """
        self.validate_feature_name(feature_name, valid_feature_names)
        self.validate_config_name(name)
        self.validate_description(description)

        return fmc.FeatureMonitoringConfig(
            feature_store_id=self._feature_store_id,
            feature_group_id=self._feature_group_id,
            feature_view_name=self._feature_view_name,
            feature_view_version=self._feature_view_version,
            name=name,
            description=description,
            feature_name=feature_name,
            # setting feature_monitoring_type to "STATISTICS_COMPARISON" allows
            # to raise an error if no reference window and comparison config are provided
            feature_monitoring_type=fmc.FeatureMonitoringType.STATISTICS_COMPARISON,
            job_schedule={
                "start_date_time": start_date_time or datetime.now(),
                "end_date_time": end_date_time,
                "cron_expression": cron_expression,
                "enabled": True,
            },
        ).with_detection_window()



#
#   Copyright 2021 Logical Clocks AB
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#

import json
from typing import Optional

from hsml.schema import Schema


class ModelSchema:
    """Create a schema for a model.

    # Arguments
        input_schema: Schema to describe the inputs.
        output_schema: Schema to describe the outputs.

    # Returns
        `ModelSchema`. The model schema object.
    """

    def __init__(
        self,
        input_schema: Optional[Schema] = None,
        output_schema: Optional[Schema] = None,
        **kwargs,
    ):
        if input_schema is not None:
            self.input_schema = input_schema

        if output_schema is not None:
            self.output_schema = output_schema

    def json(self):
        return json.dumps(
            self, default=lambda o: getattr(o, "__dict__", o), sort_keys=True, indent=2
        )

    def to_dict(self):
        """
        Get dict representation of the ModelSchema.
        """
        return json.loads(self.json())

    def __repr__(self):
        input_type = (
            self.input_schema._get_type() if hasattr(self, "input_schema") else None
        )
        output_type = (
            self.output_schema._get_type() if hasattr(self, "output_schema") else None
        )
        return f"ModelSchema(input: {input_type!r}, output: {output_type!r})"



#
#   Copyright 2024 Hopsworks AB
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
from __future__ import annotations

import datetime
import json
from dataclasses import dataclass
from enum import Enum
from typing import TYPE_CHECKING, Any, Dict, List, Optional, Union

import humps
from hopsworks_common import execution as execution_mod
from hopsworks_common import user as user_mod
from hopsworks_common import util
from hopsworks_common.core.constants import HAS_GREAT_EXPECTATIONS
from hsfs import expectation_suite as es_mod
from hsfs import validation_report as vr_mod


if TYPE_CHECKING or HAS_GREAT_EXPECTATIONS:
    from great_expectations.core import (
        ExpectationSuite,
        ExpectationSuiteValidationResult,
    )


class FeatureStoreActivityType(Enum):
    METADATA = "METADATA"
    STATISTICS = "STATISTICS"
    JOB = "JOB"
    VALIDATIONS = "VALIDATIONS"
    EXPECTATIONS = "EXPECTATIONS"
    COMMIT = "COMMIT"


@dataclass(init=False, repr=False)
class FeatureStoreActivity:
    type: FeatureStoreActivityType
    timestamp: int
    metadata: str
    user: Optional[user_mod.User]
    # optional fields depending on the activity type
    validation_report: Optional[
        Union[vr_mod.ValidationReport, ExpectationSuiteValidationResult]
    ] = None
    expectation_suite: Optional[Union[es_mod.ExpectationSuite, ExpectationSuite]] = None
    commit: Optional[Dict[str, Union[str, int, float]]] = None
    statistics: Optional[Dict[str, Union[str, int, float]]] = None
    execution: Optional[execution_mod.Execution] = None
    execution_last_event_time: Optional[int] = None
    # internal fields
    id: int
    href: str

    def __init__(
        self,
        type: str,
        timestamp: int,
        metadata: Optional[str] = None,
        user: Optional[Dict[str, Any]] = None,
        expectation_suite: Optional[Dict[str, Any]] = None,
        validation_report: Optional[Dict[str, Any]] = None,
        commit: Optional[Dict[str, Union[str, int, float]]] = None,
        statistics: Optional[Dict[str, Union[str, int, float]]] = None,
        execution: Optional[Dict[str, Any]] = None,
        execution_last_event_time: Optional[int] = None,
        **kwargs,
    ):
        self.type = FeatureStoreActivityType(type) if isinstance(type, str) else type
        self.timestamp = timestamp

        self.id = kwargs.get("id")
        self.href = kwargs.get("href")

        self.user = user_mod.User.from_response_json(user) if user else None
        self.metadata = metadata
        self.commit = commit
        self.statistics = statistics
        self.execution = (
            execution_mod.Execution.from_response_json(execution) if execution else None
        )
        self.execution_last_event_time = execution_last_event_time

        if self.type == FeatureStoreActivityType.VALIDATIONS and validation_report:
            self.validation_report = vr_mod.ValidationReport.from_response_json(
                validation_report
            )
            if HAS_GREAT_EXPECTATIONS:
                self.validation_report = self.validation_report.to_ge_type()

        if self.type == FeatureStoreActivityType.EXPECTATIONS and expectation_suite:
            self.expectation_suite = es_mod.ExpectationSuite.from_response_json(
                expectation_suite
            )
            if HAS_GREAT_EXPECTATIONS:
                self.expectation_suite = self.expectation_suite.to_ge_type()

    @classmethod
    def from_response_json(
        cls, response_json: Dict[str, Any]
    ) -> List[FeatureStoreActivity]:
        if "items" in response_json:
            return [
                cls.from_response_json(activity) for activity in response_json["items"]
            ]
        else:
            return cls(**humps.decamelize(response_json))

    def to_dict(self) -> Dict[str, Any]:
        activity_dict = {
            "id": self.id,
            "type": self.type.value,
            "metadata": self.metadata,
            "timestamp": self.timestamp,
        }
        if self.user:
            activity_dict["user"] = self.user.to_dict()
        if self.validation_report:
            activity_dict["validation_report"] = (
                self.validation_report.to_dict()
                if hasattr(self.validation_report, "_id")
                else self.validation_report.to_json_dict()
            )
        if self.expectation_suite:
            activity_dict["expectation_suite"] = (
                self.expectation_suite.to_dict()
                if hasattr(self.expectation_suite, "_id")
                else self.expectation_suite.to_json_dict()
            )
        if self.commit:
            activity_dict["commit"] = self.commit
        if self.statistics:
            activity_dict["statistics"] = self.statistics
        if self.execution:
            activity_dict["execution"] = humps.decamelize(
                json.loads(self.execution.json())
            )

        return activity_dict

    def json(self) -> str:
        return json.dumps(self.to_dict(), cls=util.Encoder)

    def __repr__(self):
        utc_human_readable = (
            datetime.datetime.fromtimestamp(
                self.timestamp / 1000, datetime.timezone.utc
            ).strftime(r"%Y-%m-%d %H:%M:%S")
            + " UTC"
        )
        the_string = f"Activity {self.type.value},"
        the_string += f" at: {utc_human_readable}"
        if self.user:
            the_string += f", by: {self.user.email}"
        if self.metadata:
            the_string += f"\n\t{self.metadata}"
        if self.execution:
            the_string += f"\n{self.execution.get_url()},"
        if self.validation_report:
            the_string += f"Validation {'succeeded' if self.validation_report.success else 'failed'}."
        if self.expectation_suite:
            the_string += (
                f"It has {len(self.expectation_suite.expectations)} expectations."
            )
        if self.statistics:
            the_string += f"\nComputed following statistics:\n{json.dumps(self.statistics, indent=2)}"
        if self.commit:
            the_string += f"\nData ingestion:\n{json.dumps(self.commit, indent=2)}"

        return the_string



#
#   Copyright 2021 Logical Clocks AB
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#

from __future__ import annotations

import datetime
import inspect
import os
import shutil
from json import JSONEncoder, dumps
from urllib.parse import urljoin, urlparse

import humps
import numpy as np
import pandas as pd
from hsml import client
from hsml.constants import MODEL, PREDICTOR, Default
from six import string_types


class VersionWarning(Warning):
    pass


class ProvenanceWarning(Warning):
    pass


class MLEncoder(JSONEncoder):
    def default(self, obj):
        try:
            return obj.to_dict()
        except AttributeError:
            return super().default(obj)


class NumpyEncoder(JSONEncoder):
    """Special json encoder for numpy types.
    Note that some numpy types doesn't have native python equivalence,
    hence json.dumps will raise TypeError.
    In this case, you'll need to convert your numpy types into its closest python equivalence.
    """

    def convert(self, obj):
        import base64

        import numpy as np
        import pandas as pd

        def encode_binary(x):
            return base64.encodebytes(x).decode("ascii")

        if isinstance(obj, np.ndarray):
            if obj.dtype == np.object:
                return [self.convert(x)[0] for x in obj.tolist()]
            elif obj.dtype == np.bytes_:
                return np.vectorize(encode_binary)(obj), True
            else:
                return obj.tolist(), True

        if isinstance(obj, (pd.Timestamp, datetime.date)):
            return obj.isoformat(), True
        if isinstance(obj, bytes) or isinstance(obj, bytearray):
            return encode_binary(obj), True
        if isinstance(obj, np.generic):
            return obj.item(), True
        if isinstance(obj, np.datetime64):
            return np.datetime_as_string(obj), True
        return obj, False

    def default(self, obj):  # pylint: disable=E0202
        res, converted = self.convert(obj)
        if converted:
            return res
        else:
            return super().default(obj)


# Model registry

# - schema and types


def set_model_class(model):
    from hsml.llm.model import Model as LLMModel
    from hsml.model import Model as BaseModel
    from hsml.python.model import Model as PyModel
    from hsml.sklearn.model import Model as SkLearnModel
    from hsml.tensorflow.model import Model as TFModel
    from hsml.torch.model import Model as TorchModel

    if "href" in model:
        _ = model.pop("href")
    if "type" in model:  # backwards compatibility
        _ = model.pop("type")
    if "tags" in model:
        _ = model.pop("tags")  # tags are always retrieved from backend

    if "framework" not in model:
        return BaseModel(**model)

    framework = model.pop("framework")
    if framework == MODEL.FRAMEWORK_TENSORFLOW:
        return TFModel(**model)
    if framework == MODEL.FRAMEWORK_TORCH:
        return TorchModel(**model)
    if framework == MODEL.FRAMEWORK_SKLEARN:
        return SkLearnModel(**model)
    elif framework == MODEL.FRAMEWORK_PYTHON:
        return PyModel(**model)
    elif framework == MODEL.FRAMEWORK_LLM:
        return LLMModel(**model)
    else:
        raise ValueError(
            "framework {} is not a supported framework".format(str(framework))
        )


def input_example_to_json(input_example):
    if isinstance(input_example, np.ndarray):
        if input_example.size > 0:
            return _handle_tensor_input(input_example)
        else:
            raise ValueError(
                "input_example of type {} can not be empty".format(type(input_example))
            )
    elif isinstance(input_example, dict):
        return _handle_dict_input(input_example)
    else:
        return _handle_dataframe_input(input_example)


def _handle_tensor_input(input_tensor):
    return input_tensor.tolist()


def _handle_dataframe_input(input_ex):
    if isinstance(input_ex, pd.DataFrame):
        if not input_ex.empty:
            return input_ex.iloc[0].tolist()
        else:
            raise ValueError(
                "input_example of type {} can not be empty".format(type(input_ex))
            )
    elif isinstance(input_ex, pd.Series):
        if not input_ex.empty:
            return input_ex.tolist()
        else:
            raise ValueError(
                "input_example of type {} can not be empty".format(type(input_ex))
            )
    elif isinstance(input_ex, list):
        if len(input_ex) > 0:
            return input_ex
        else:
            raise ValueError(
                "input_example of type {} can not be empty".format(type(input_ex))
            )
    else:
        raise TypeError(
            "{} is not a supported input example type".format(type(input_ex))
        )


def _handle_dict_input(input_ex):
    return input_ex


# - artifacts


def compress(archive_out_path, archive_name, path_to_archive):
    if os.path.isdir(path_to_archive):
        return shutil.make_archive(
            os.path.join(archive_out_path, archive_name), "gztar", path_to_archive
        )
    else:
        return shutil.make_archive(
            os.path.join(archive_out_path, archive_name),
            "gztar",
            os.path.dirname(path_to_archive),
            os.path.basename(path_to_archive),
        )


def decompress(archive_file_path, extract_dir=None):
    return shutil.unpack_archive(archive_file_path, extract_dir=extract_dir)


# - export models


def validate_metrics(metrics):
    if metrics is not None:
        if not isinstance(metrics, dict):
            raise TypeError(
                "provided metrics is of instance {}, expected a dict".format(
                    type(metrics)
                )
            )

        for metric in metrics:
            # Validate key is a string
            if not isinstance(metric, string_types):
                raise TypeError(
                    "provided metrics key is of instance {}, expected a string".format(
                        type(metric)
                    )
                )
            # Validate value is a number
            try:
                float(metrics[metric])
            except ValueError as err:
                raise ValueError(
                    "{} is not a number, only numbers can be attached as metadata for models.".format(
                        str(metrics[metric])
                    )
                ) from err


# Model serving


def get_predictor_for_model(model, **kwargs):
    from hsml.llm.model import Model as LLMModel
    from hsml.llm.predictor import Predictor as vLLMPredictor
    from hsml.model import Model as BaseModel
    from hsml.predictor import Predictor as BasePredictor
    from hsml.python.model import Model as PyModel
    from hsml.python.predictor import Predictor as PyPredictor
    from hsml.sklearn.model import Model as SkLearnModel
    from hsml.sklearn.predictor import Predictor as SkLearnPredictor
    from hsml.tensorflow.model import Model as TFModel
    from hsml.tensorflow.predictor import Predictor as TFPredictor
    from hsml.torch.model import Model as TorchModel
    from hsml.torch.predictor import Predictor as TorchPredictor

    if not isinstance(model, BaseModel):
        raise ValueError(
            "model is of type {}, but an instance of {} class is expected".format(
                type(model), BaseModel
            )
        )

    if type(model) is TFModel:
        return TFPredictor(**kwargs)
    if type(model) is TorchModel:
        return TorchPredictor(**kwargs)
    if type(model) is SkLearnModel:
        return SkLearnPredictor(**kwargs)
    if type(model) is PyModel:
        return PyPredictor(**kwargs)
    if type(model) is LLMModel:
        return vLLMPredictor(**kwargs)
    if type(model) is BaseModel:
        return BasePredictor(  # python as default framework and model server
            model_framework=MODEL.FRAMEWORK_PYTHON,
            model_server=PREDICTOR.MODEL_SERVER_PYTHON,
            **kwargs,
        )


def get_hostname_replaced_url(sub_path: str):
    """
    construct and return an url with public hopsworks hostname and sub path
    :param self:
    :param sub_path: url sub-path after base url
    :return: href url
    """
    href = urljoin(client.get_instance()._base_url, sub_path)
    url_parsed = client.get_instance().replace_public_host(urlparse(href))
    return url_parsed.geturl()


# General


def pretty_print(obj):
    if isinstance(obj, list):
        for logs in obj:
            pretty_print(logs)
    else:
        json_decamelized = humps.decamelize(obj.to_dict())
        print(dumps(json_decamelized, indent=4, sort_keys=True))


def get_members(cls, prefix=None):
    for m in inspect.getmembers(cls, lambda m: not (inspect.isroutine(m))):
        n = m[0]  # name
        if (prefix is not None and n.startswith(prefix)) or (
            prefix is None and not (n.startswith("__") and n.endswith("__"))
        ):
            yield m[1]  # value


# - json


def extract_field_from_json(obj, fields, default=None, as_instance_of=None):
    if isinstance(fields, list):
        for field in fields:
            value = extract_field_from_json(obj, field, default, as_instance_of)
            if value is not None:
                break
    else:
        value = obj.pop(fields) if fields in obj else default
        if as_instance_of is not None:
            if isinstance(value, list):
                # if the field is a list, get all obj
                value = [
                    get_obj_from_json(obj=subvalue, cls=as_instance_of)
                    for subvalue in value
                ]
            else:
                # otherwise, get single obj
                value = get_obj_from_json(obj=value, cls=as_instance_of)
    return value


def get_obj_from_json(obj, cls):
    if obj is not None:
        if isinstance(obj, cls):
            return obj
        if isinstance(obj, dict):
            return cls.from_json(obj)
        if isinstance(obj, Default):
            return cls()
        raise ValueError(
            "Object of type {} cannot be converted to class {}".format(type(obj), cls)
        )
    return obj


def feature_view_to_json(obj):
    if obj is None:
        return None
    import importlib.util

    if importlib.util.find_spec("hsfs"):
        from hsfs import feature_view

        if isinstance(obj, feature_view.FeatureView):
            import json

            import humps

            return humps.camelize(json.loads(obj.json()))
    return None



#
#   Copyright 2020 Logical Clocks AB
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#

from __future__ import annotations

import os

import requests
from hopsworks_common.client import exceptions


class BearerAuth(requests.auth.AuthBase):
    """Class to encapsulate a Bearer token."""

    def __init__(self, token: str) -> None:
        self._token = token.strip()

    def __call__(self, r: requests.Request) -> requests.Request:
        r.headers["Authorization"] = "Bearer " + self._token
        return r


class ApiKeyAuth(requests.auth.AuthBase):
    """Class to encapsulate an API key."""

    def __init__(self, token: str) -> None:
        self._token = token.strip()

    def __call__(self, r: requests.Request) -> requests.Request:
        r.headers["Authorization"] = "ApiKey " + self._token
        return r


class OnlineStoreKeyAuth(requests.auth.AuthBase):
    """Class to encapsulate an API key."""

    def __init__(self, token):
        self._token = token.strip()

    def __call__(self, r):
        r.headers["X-API-KEY"] = self._token
        return r


def get_api_key(api_key_value, api_key_file):
    if api_key_value is not None:
        return api_key_value
    elif api_key_file is not None:
        file = None
        if os.path.exists(api_key_file):
            try:
                file = open(api_key_file, mode="r")
                return file.read()
            finally:
                file.close()
        else:
            raise IOError(
                "Could not find api key file on path: {}".format(api_key_file)
            )
    else:
        raise exceptions.ExternalClientError(
            "Either api_key_file or api_key_value must be set when connecting to"
            " hopsworks from an external environment."
        )



#
#   Copyright 2020 Logical Clocks AB
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
from __future__ import annotations

import importlib
import importlib.util
import os
import re
import sys
import warnings
import weakref
from typing import Any, Optional

from hopsworks_common import client, usage, util, version
from hopsworks_common.core import (
    hosts_api,
    project_api,
    secret_api,
    services_api,
    variable_api,
)
from hopsworks_common.core.opensearch import OpenSearchClientSingleton
from hopsworks_common.decorators import connected, not_connected
from requests.exceptions import ConnectionError


HOPSWORKS_PORT_DEFAULT = 443
HOSTNAME_VERIFICATION_DEFAULT = os.environ.get(
    "HOPSWORKS_HOSTNAME_VERIFICATION", "False"
).lower() in ("true", "1", "y", "yes")
# alias for backwards compatibility:
HOPSWORKS_HOSTNAME_VERIFICATION_DEFAULT = HOSTNAME_VERIFICATION_DEFAULT
CERT_FOLDER_DEFAULT = "/tmp"
PROJECT_ID = "HOPSWORKS_PROJECT_ID"
PROJECT_NAME = "HOPSWORKS_PROJECT_NAME"


_hsfs_engine_type = None


class Connection:
    """A hopsworks connection object.

    This class provides convenience classmethods accessible from the `hopsworks`-module:

    !!! example "Connection factory"
        For convenience, `hopsworks` provides a factory method, accessible from the top level
        module, so you don't have to import the `Connection` class manually:

        ```python
        import hopsworks
        conn = hopsworks.connection()
        ```

    !!! hint "Save API Key as File"
        To get started quickly, you can simply create a file with the previously
         created Hopsworks API Key and place it on the environment from which you
         wish to connect to Hopsworks.

        You can then connect by simply passing the path to the key file when
        instantiating a connection:

        ```python hl_lines="6"
            import hopsworks
            conn = hopsworks.connection(
                'my_instance',                      # DNS of your Hopsworks instance
                443,                                # Port to reach your Hopsworks instance, defaults to 443
                api_key_file='hopsworks.key',       # The file containing the API key generated above
                hostname_verification=True)         # Disable for self-signed certificates
            )
            project = conn.get_project("my_project")
        ```

    Clients in external clusters need to connect to the Hopsworks using an
    API key. The API key is generated inside the Hopsworks platform, and requires at
    least the "project" scope to be able to access a project.
    For more information, see the [integration guides](../setup.md).

    # Arguments
        host: The hostname of the Hopsworks instance in the form of `[UUID].cloud.hopsworks.ai`,
            defaults to `None`. Do **not** use the url including `https://` when connecting
            programatically.
        port: The port on which the Hopsworks instance can be reached,
            defaults to `443`.
        project: The name of the project to connect to. When running on Hopsworks, this
            defaults to the project from where the client is run from.
            Defaults to `None`.
        engine: Which engine to use, `"spark"`, `"python"` or `"training"`. Defaults to `None`,
            which initializes the engine to Spark if the environment provides Spark, for
            example on Hopsworks and Databricks, or falls back on Hive in Python if Spark is not
            available, e.g. on local Python environments or AWS SageMaker. This option
            allows you to override this behaviour. `"training"` engine is useful when only
            feature store metadata is needed, for example training dataset location and label
            information when Hopsworks training experiment is conducted.
        hostname_verification: Whether or not to verify Hopsworks' certificate, defaults
            to `True`.
        trust_store_path: Path on the file system containing the Hopsworks certificates,
            defaults to `None`.
        cert_folder: The directory to store retrieved HopsFS certificates, defaults to
            `"/tmp"`. Only required when running without a Spark environment.
        api_key_file: Path to a file containing the API Key, defaults to `None`.
        api_key_value: API Key as string, if provided, `api_key_file` will be ignored,
            however, this should be used with care, especially if the used notebook or
            job script is accessible by multiple parties. Defaults to `None`.

    # Returns
        `Connection`. Connection handle to perform operations on a
            Hopsworks project.
    """

    def __init__(
        self,
        host: Optional[str] = None,
        port: int = HOPSWORKS_PORT_DEFAULT,
        project: Optional[str] = None,
        engine: Optional[str] = None,
        hostname_verification: bool = HOSTNAME_VERIFICATION_DEFAULT,
        trust_store_path: Optional[str] = None,
        cert_folder: str = CERT_FOLDER_DEFAULT,
        api_key_file: Optional[str] = None,
        api_key_value: Optional[str] = None,
    ) -> None:
        self._host = host
        self._port = port
        self._project = project
        self._engine = engine
        self._hostname_verification = hostname_verification
        self._trust_store_path = trust_store_path
        self._cert_folder = cert_folder
        self._api_key_file = api_key_file
        self._api_key_value = api_key_value
        self._connected = False

        self.connect()

    @usage.method_logger
    @connected
    def get_feature_store(
        self,
        name: Optional[str] = None,
        engine: Optional[str] = None,
    ):  # -> feature_store.FeatureStore
        # the typing is commented out due to circular dependency, it breaks auto_doc.py
        """Get a reference to a feature store to perform operations on.

        Defaulting to the project name of default feature store. To get a
        Shared feature stores, the project name of the feature store is required.

        # Arguments
            name: The name of the feature store, defaults to `None`.
            engine: Which engine to use, `"spark"`, `"python"` or `"training"`. Defaults to `None`,
            which initializes the engine to Spark if the environment provides Spark, for
            example on Hopsworks and Databricks, or falls back on Hive in Python if Spark is not
            available, e.g. on local Python environments or AWS SageMaker. This option
            allows you to override this behaviour. `"training"` engine is useful when only
            feature store metadata is needed, for example training dataset location and label
            information when Hopsworks training experiment is conducted.

        # Returns
            `FeatureStore`. A feature store handle object to perform operations on.
        """
        # Ensure the engine is initialized and of right type
        from hsfs import engine as hsfs_engine

        if engine:
            global _hsfs_engine_type
            _hsfs_engine_type = engine
        hsfs_engine.get_instance()

        if not name:
            name = client.get_instance()._project_name
        return self._feature_store_api.get(util.append_feature_store_suffix(name))

    @usage.method_logger
    @connected
    def get_model_registry(self, project: str = None):
        """Get a reference to a model registry to perform operations on, defaulting to the project's default model registry.
        Shared model registries can be retrieved by passing the `project` argument.

        # Arguments
            project: The name of the project that owns the shared model registry,
            the model registry must be shared with the project the connection was established for, defaults to `None`.
        # Returns
            `ModelRegistry`. A model registry handle object to perform operations on.
        """
        return self._model_registry_api.get(project)

    @usage.method_logger
    @connected
    def get_model_serving(self):
        """Get a reference to model serving to perform operations on. Model serving operates on top of a model registry, defaulting to the project's default model registry.

        !!! example
            ```python

            import hopsworks

            project = hopsworks.login()

            ms = project.get_model_serving()
            ```

        # Returns
            `ModelServing`. A model serving handle object to perform operations on.
        """
        return self._model_serving_api.get()

    @usage.method_logger
    @connected
    def get_secrets_api(self):
        """Get the secrets api.

        # Returns
            `SecretsApi`: The Secrets Api handle
        """
        return self._secret_api

    @usage.method_logger
    @connected
    def create_project(
        self, name: str, description: str = None, feature_store_topic: str = None
    ):
        """Create a new project.

        Example for creating a new project

        ```python

        import hopsworks

        connection = hopsworks.connection()

        connection.create_project("my_hopsworks_project", description="An example Hopsworks project")

        ```
        # Arguments
            name: The name of the project.
            description: optional description of the project
            feature_store_topic: optional feature store topic name

        # Returns
            `Project`. A project handle object to perform operations on.
        """
        return self._project_api._create_project(name, description, feature_store_topic)

    @usage.method_logger
    @connected
    def get_project(self, name: str = None):
        """Get an existing project.

        # Arguments
            name: The name of the project.

        # Returns
            `Project`. A project handle object to perform operations on.
        """
        _client = client.get_instance()
        if not name and not _client._project_name:
            raise ValueError(
                "No project name provided. Please provide a project name or"
                " set a project when login or creating the connection."
            )
        elif not _client._project_name:
            self._provide_project(name)
        elif not name:
            name = client.get_instance()._project_name

        return self._project_api._get_project(name)

    @usage.method_logger
    @connected
    def get_projects(self):
        """Get all projects.

        # Returns
            `List[Project]`: List of Project objects
        """

        return self._project_api._get_projects()

    @usage.method_logger
    @connected
    def project_exists(self, name: str):
        """Check if a project exists.

        # Arguments
            name: The name of the project.

        # Returns
            `bool`. True if project exists, otherwise False
        """
        return self._project_api._exists(name)

    @connected
    def _check_compatibility(self):
        """Check the compatibility between the client and backend.
        Assumes versioning (major.minor.patch).
        A client is considered compatible if the major and minor version matches.

        """

        versionPattern = r"\d+\.\d+"
        regexMatcher = re.compile(versionPattern)

        client_version = version.__version__
        backend_version = self._variable_api.get_version("hopsworks")

        major_minor_client = regexMatcher.search(client_version).group(0)
        major_minor_backend = regexMatcher.search(backend_version).group(0)

        if major_minor_backend != major_minor_client:
            print("\n", file=sys.stderr)
            warnings.warn(
                "The installed hopsworks client version {0} may not be compatible with the connected Hopsworks backend version {1}. \nTo ensure compatibility please install the latest bug fix release matching the minor version of your backend ({2}) by running 'pip install hopsworks=={2}.*'".format(
                    client_version, backend_version, major_minor_backend
                ),
                stacklevel=1,
            )
            sys.stderr.flush()

    @not_connected
    def connect(self) -> None:
        """Instantiate the connection.

        Creating a `Connection` object implicitly calls this method for you to
        instantiate the connection. However, it is possible to close the connection
        gracefully with the `close()` method, in order to clean up materialized
        certificates. This might be desired when working on external environments such
        as AWS SageMaker. Subsequently you can call `connect()` again to reopen the
        connection.

        !!! example
            ```python
            import hopsworks
            conn = hopsworks.connection()
            conn.close()
            conn.connect()
            ```
        """
        client.stop()
        self._connected = True
        finalizer = weakref.finalize(self, self.close)
        try:
            # determine engine, needed to init client
            if (self._engine is not None and self._engine.lower() == "spark") or (
                self._engine is None and importlib.util.find_spec("pyspark")
            ):
                self._engine = "spark"
            elif (self._engine is not None and self._engine.lower() == "python") or (
                self._engine is None and not importlib.util.find_spec("pyspark")
            ):
                self._engine = "python"
            elif self._engine is not None and self._engine.lower() == "training":
                self._engine = "training"
            elif (
                self._engine is not None
                and self._engine.lower() == "spark-no-metastore"
            ):
                self._engine = "spark-no-metastore"
            else:
                raise ConnectionError(
                    "Engine you are trying to initialize is unknown. "
                    "Supported engines are `'spark'`, `'python'` and `'training'`."
                )

            # init client
            if client.base.Client.REST_ENDPOINT not in os.environ:
                client.init(
                    "external",
                    self._host,
                    self._port,
                    self._project,
                    self._engine,
                    self._hostname_verification,
                    self._trust_store_path,
                    self._cert_folder,
                    self._api_key_file,
                    self._api_key_value,
                )
            else:
                client.init(
                    "hopsworks",
                    hostname_verification=self._hostname_verification,
                )

            client.set_connection(self)
            from hsfs.core import feature_store_api
            from hsml.core import model_registry_api, model_serving_api

            global _hsfs_engine_type
            _hsfs_engine_type = self._engine
            self._feature_store_api = feature_store_api.FeatureStoreApi()
            self._model_registry_api = model_registry_api.ModelRegistryApi()
            self._model_serving_api = model_serving_api.ModelServingApi()
            self._project_api = project_api.ProjectApi()
            self._hosts_api = hosts_api.HostsApi()
            self._services_api = services_api.ServicesApi()
            self._secret_api = secret_api.SecretsApi()
            self._variable_api = variable_api.VariableApi()
            usage.init_usage(self._host, self._variable_api.get_version("hopsworks"))

            self._provide_project()
        except (TypeError, ConnectionError):
            self._connected = False
            finalizer.detach()
            raise

        self._check_compatibility()

    @connected
    def _provide_project(self, name=None):
        _client = client.get_instance()

        if name:
            self._project = name
            if _client._is_external():
                _client.provide_project(name)

        if _client._project_name:
            self._project = _client._project_name

        if not self._project:
            return

        from hsfs import engine

        engine.get_instance()
        if self._variable_api.get_data_science_profile_enabled():
            # load_default_configuration has to be called before using hsml
            # but after a project is provided to client
            self._model_serving_api.load_default_configuration()  # istio client, default resources,...

    def close(self) -> None:
        """Close a connection gracefully.

        This will clean up any materialized certificates on the local file system of
        external environments such as AWS SageMaker.

        Usage is optional.

        !!! example
            ```python
            import hopsworks
            conn = hopsworks.connection()
            conn.close()
            ```
        """
        if not self._connected:
            return  # the connection is already closed

        from hsfs import engine

        if OpenSearchClientSingleton._instance:
            OpenSearchClientSingleton().close()
        client.stop()
        engine.stop()
        self._feature_store_api = None
        self._connected = False
        print("Connection closed.")

    @classmethod
    def connection(
        cls,
        host: Optional[str] = None,
        port: int = HOPSWORKS_PORT_DEFAULT,
        project: Optional[str] = None,
        engine: Optional[str] = None,
        hostname_verification: bool = HOSTNAME_VERIFICATION_DEFAULT,
        trust_store_path: Optional[str] = None,
        cert_folder: str = CERT_FOLDER_DEFAULT,
        api_key_file: Optional[str] = None,
        api_key_value: Optional[str] = None,
    ) -> Connection:
        """Connection factory method, accessible through `hopsworks.connection()`.

        This class provides convenience classmethods accessible from the `hopsworks`-module:

        !!! example "Connection factory"
            For convenience, `hopsworks` provides a factory method, accessible from the top level
            module, so you don't have to import the `Connection` class manually:

            ```python
            import hopsworks
            conn = hopsworks.connection()
            ```

        !!! hint "Save API Key as File"
            To get started quickly, you can simply create a file with the previously
            created Hopsworks API Key and place it on the environment from which you
            wish to connect to Hopsworks.

            You can then connect by simply passing the path to the key file when
            instantiating a connection:

            ```python hl_lines="6"
                import hopsworks
                conn = hopsworks.connection(
                    'my_instance',                      # DNS of your Hopsworks instance
                    443,                                # Port to reach your Hopsworks instance, defaults to 443
                    api_key_file='hopsworks.key',       # The file containing the API key generated above
                    hostname_verification=True)         # Disable for self-signed certificates
                )
                project = conn.get_project("my_project")
            ```

        Clients in external clusters need to connect to the Hopsworks using an
        API key. The API key is generated inside the Hopsworks platform, and requires at
        least the "project" scope to be able to access a project.
        For more information, see the [integration guides](../setup.md).

        # Arguments
            host: The hostname of the Hopsworks instance in the form of `[UUID].cloud.hopsworks.ai`,
                defaults to `None`. Do **not** use the url including `https://` when connecting
                programatically.
            port: The port on which the Hopsworks instance can be reached,
                defaults to `443`.
            project: The name of the project to connect to. When running on Hopsworks, this
                defaults to the project from where the client is run from.
                Defaults to `None`.
            engine: Which engine to use, `"spark"`, `"python"` or `"training"`. Defaults to `None`,
                which initializes the engine to Spark if the environment provides Spark, for
                example on Hopsworks and Databricks, or falls back on Hive in Python if Spark is not
                available, e.g. on local Python environments or AWS SageMaker. This option
                allows you to override this behaviour. `"training"` engine is useful when only
                feature store metadata is needed, for example training dataset location and label
                information when Hopsworks training experiment is conducted.
            hostname_verification: Whether or not to verify Hopsworks' certificate, defaults
                to `True`.
            trust_store_path: Path on the file system containing the Hopsworks certificates,
                defaults to `None`.
            cert_folder: The directory to store retrieved HopsFS certificates, defaults to
                `"/tmp"`. Only required when running without a Spark environment.
            api_key_file: Path to a file containing the API Key, defaults to `None`.
            api_key_value: API Key as string, if provided, `api_key_file` will be ignored,
                however, this should be used with care, especially if the used notebook or
                job script is accessible by multiple parties. Defaults to `None`.

        # Returns
            `Connection`. Connection handle to perform operations on a
                Hopsworks project.
        """
        return cls(
            host,
            port,
            project,
            engine,
            hostname_verification,
            trust_store_path,
            cert_folder,
            api_key_file,
            api_key_value,
        )

    @property
    def host(self) -> Optional[str]:
        return self._host

    @host.setter
    @not_connected
    def host(self, host: Optional[str]) -> None:
        self._host = host

    @property
    def port(self) -> int:
        return self._port

    @port.setter
    @not_connected
    def port(self, port) -> int:
        self._port = port

    @property
    def project(self) -> Optional[str]:
        return self._project

    @project.setter
    @not_connected
    def project(self, project: Optional[str]) -> str:
        self._project = project

    @property
    def hostname_verification(self):
        return self._hostname_verification

    @hostname_verification.setter
    @not_connected
    def hostname_verification(self, hostname_verification):
        self._hostname_verification = hostname_verification

    @property
    def trust_store_path(self) -> Optional[str]:
        return self._trust_store_path

    @trust_store_path.setter
    @not_connected
    def trust_store_path(self, trust_store_path: Optional[str]) -> None:
        self._trust_store_path = trust_store_path

    @property
    def cert_folder(self) -> str:
        return self._cert_folder

    @cert_folder.setter
    @not_connected
    def cert_folder(self, cert_folder: str) -> None:
        self._cert_folder = cert_folder

    @property
    def api_key_file(self) -> Optional[str]:
        return self._api_key_file

    @property
    def api_key_value(self) -> Optional[str]:
        return self._api_key_value

    @api_key_file.setter
    @not_connected
    def api_key_file(self, api_key_file: Optional[str]) -> None:
        self._api_key_file = api_key_file

    @api_key_value.setter
    @not_connected
    def api_key_value(self, api_key_value: Optional[str]) -> Optional[str]:
        self._api_key_value = api_key_value

    def __enter__(self) -> Connection:
        self.connect()
        return self

    def __exit__(self, type: Any, value: Any, traceback: Any):
        self.close()



#
#   Copyright 2020 Logical Clocks AB
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
from __future__ import annotations

from hopsworks_common.client.exceptions import (
    DataValidationException,
    FeatureStoreException,
)
from hsfs import engine, util
from hsfs import feature_group as fg
from hsfs.core import feature_group_base_engine


class ExternalFeatureGroupEngine(feature_group_base_engine.FeatureGroupBaseEngine):
    def save(self, feature_group):
        if feature_group.features is None or len(feature_group.features) == 0:
            # If the user didn't specify the schema, parse it from the query
            external_dataset = engine.get_instance().register_external_temporary_table(
                feature_group, "read_ondmd"
            )
            # if python engine user should pass features as we do not parse it in this case
            if external_dataset is None:
                raise FeatureStoreException(
                    "Features (schema) need to be set for creation of external feature groups with engine "
                    + engine.get_type()
                    + ". Alternatively use Spark kernel."
                )

            feature_group._features = engine.get_instance().parse_schema_feature_group(
                external_dataset
            )

        # set primary and partition key columns
        # we should move this to the backend
        util.verify_attribute_key_names(feature_group, True)
        for feat in feature_group.features:
            if feat.name in feature_group.primary_key:
                feat.primary = True
        util.validate_embedding_feature_type(
            feature_group.embedding_index, feature_group._features
        )

        self._feature_group_api.save(feature_group)

    def insert(
        self,
        feature_group,
        feature_dataframe,
        write_options: dict,
        validation_options: dict = None,
    ):
        if not feature_group.online_enabled:
            raise FeatureStoreException(
                "Online storage is not enabled for this feature group. External feature groups can only store data in"
                + " online storage. To create an offline only external feature group, use the `save` method."
            )

        schema = engine.get_instance().parse_schema_feature_group(feature_dataframe)
        util.validate_embedding_feature_type(feature_group.embedding_index, schema)

        if not feature_group._id:
            # only save metadata if feature group does not exist
            feature_group.features = schema
            self.save(feature_group)
        else:
            # else, just verify that feature group schema matches user-provided dataframe
            self._verify_schema_compatibility(feature_group.features, schema)

        # ge validation on python and non stream feature groups on spark
        ge_report = feature_group._great_expectation_engine.validate(
            feature_group=feature_group,
            dataframe=feature_dataframe,
            validation_options=validation_options or {},
            ingestion_result="INGESTED",
            ge_type=False,
        )

        if ge_report is not None and ge_report.ingestion_result == "REJECTED":
            feature_group_url = util.get_feature_group_url(
                feature_store_id=feature_group.feature_store_id,
                feature_group_id=feature_group.id,
            )
            raise DataValidationException(
                "Data validation failed while validation ingestion policy set to strict, "
                + f"insertion to {feature_group.name} was aborted.\n"
                + f"You can check a summary or download your report at {feature_group_url}"
            )

        return (
            engine.get_instance().save_dataframe(
                feature_group=feature_group,
                dataframe=feature_dataframe,
                operation=None,
                online_enabled=feature_group.online_enabled,
                storage="online",
                offline_write_options=write_options,
                online_write_options=write_options,
            ),
            ge_report,
        )

    def _update_features_metadata(self, feature_group, features):
        # perform changes on copy in case the update fails, so we don't leave
        # the user object in corrupted state
        fg_dict = feature_group.to_dict()
        copy_feature_group = fg.ExternalFeatureGroup.from_response_json(fg_dict)
        copy_feature_group.features = features
        self._feature_group_api.update_metadata(
            feature_group, copy_feature_group, "updateMetadata"
        )

    def update_features(self, feature_group, updated_features):
        """Updates features safely."""
        self._update_features_metadata(
            feature_group, self.new_feature_list(feature_group, updated_features)
        )

    def append_features(self, feature_group, new_features):
        """Appends features to a feature group."""
        self._update_features_metadata(
            feature_group,
            feature_group.features + new_features,  # todo allows for duplicates
        )

    def update_description(self, feature_group, description):
        """Updates the description of a feature group."""
        fg_dict = feature_group.to_dict()
        copy_feature_group = fg.ExternalFeatureGroup.from_response_json(fg_dict)
        copy_feature_group.description = description
        self._feature_group_api.update_metadata(
            feature_group, copy_feature_group, "updateMetadata"
        )

    def update_deprecated(self, feature_group, deprecate):
        """Updates the deprecation status of a feature group."""
        fg_dict = feature_group.to_dict()
        copy_feature_group = fg.ExternalFeatureGroup.from_response_json(fg_dict)
        self._feature_group_api.update_metadata(
            feature_group, copy_feature_group, "deprecate", deprecate
        )



#
#   Copyright 2024 Hopsworks AB
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

import grpc
from hopsworks_common.client.istio.grpc.proto.grpc_predict_v2_pb2_grpc import (
    GRPCInferenceServiceStub,
)
from hopsworks_common.client.istio.utils.infer_type import InferRequest, InferResponse


class GRPCInferenceServerClient:
    def __init__(
        self,
        url,
        serving_api_key,
        channel_args=None,
    ):
        if channel_args is not None:
            channel_opt = channel_args
        else:
            channel_opt = [
                ("grpc.max_send_message_length", -1),
                ("grpc.max_receive_message_length", -1),
            ]

        # Authentication is done via API Key in the Authorization header
        self._channel = grpc.insecure_channel(url, options=channel_opt)
        self._client_stub = GRPCInferenceServiceStub(self._channel)
        self._serving_api_key = serving_api_key

    def __enter__(self):
        return self

    def __exit__(self, type, value, traceback):
        self.close()

    def __del__(self):
        """It is called during object garbage collection."""
        self.close()

    def close(self):
        """Close the client. Future calls to server will result in an Error."""
        self._channel.close()

    def infer(self, infer_request: InferRequest, headers=None, client_timeout=None):
        headers = {} if headers is None else headers
        headers["authorization"] = "ApiKey " + self._serving_api_key
        metadata = headers.items()

        # convert the InferRequest to a ModelInferRequest message
        request = infer_request.to_grpc()

        try:
            # send request
            model_infer_response = self._client_stub.ModelInfer(
                request=request, metadata=metadata, timeout=client_timeout
            )
        except grpc.RpcError as rpc_error:
            raise rpc_error

        # convert back the ModelInferResponse message to InferResponse
        return InferResponse.from_grpc(model_infer_response)



#
#   Copyright 2024 Hopsworks AB
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#

from hopsworks_common.core.type_systems import (
    PYARROW_HOPSWORKS_DTYPE_MAPPING,
    cast_column_to_offline_type,
    cast_column_to_online_type,
    cast_pandas_column_to_offline_type,
    cast_polars_column_to_offline_type,
    convert_pandas_dtype_to_offline_type,
    convert_pandas_object_type_to_offline_type,
    convert_simple_pandas_dtype_to_offline_type,
    convert_spark_type_to_offline_type,
    infer_spark_type,
    translate_legacy_spark_type,
)


__all__ = [
    "PYARROW_HOPSWORKS_DTYPE_MAPPING",
    "cast_column_to_offline_type",
    "cast_column_to_online_type",
    "cast_pandas_column_to_offline_type",
    "cast_polars_column_to_offline_type",
    "convert_pandas_dtype_to_offline_type",
    "convert_pandas_object_type_to_offline_type",
    "convert_simple_pandas_dtype_to_offline_type",
    "convert_spark_type_to_offline_type",
    "infer_spark_type",
    "translate_legacy_spark_type",
]



#
#   Copyright 2022 Logical Clocks AB
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

import json
from typing import Optional

import humps
from hsml import util
from hsml.constants import KAFKA_TOPIC


class KafkaTopic:
    """Configuration for a Kafka topic."""

    def __init__(
        self,
        name: str = KAFKA_TOPIC.CREATE,
        num_replicas: Optional[int] = None,
        num_partitions: Optional[int] = None,
        **kwargs,
    ):
        self._name = name
        self._num_replicas, self._num_partitions = self._validate_topic_config(
            self._name, num_replicas, num_partitions
        )

    def describe(self):
        util.pretty_print(self)

    @classmethod
    def _validate_topic_config(cls, name, num_replicas, num_partitions):
        if name is not None and name != KAFKA_TOPIC.NONE:
            if name == KAFKA_TOPIC.CREATE:
                if num_replicas is None:
                    print(
                        "Setting number of replicas to default value '{}'".format(
                            KAFKA_TOPIC.NUM_REPLICAS
                        )
                    )
                    num_replicas = KAFKA_TOPIC.NUM_REPLICAS
                if num_partitions is None:
                    print(
                        "Setting number of partitions to default value '{}'".format(
                            KAFKA_TOPIC.NUM_PARTITIONS
                        )
                    )
                    num_partitions = KAFKA_TOPIC.NUM_PARTITIONS
            else:
                if num_replicas is not None or num_partitions is not None:
                    raise ValueError(
                        "Number of replicas or partitions cannot be changed in existing kafka topics."
                    )
        elif name is None or name == KAFKA_TOPIC.NONE:
            num_replicas = None
            num_partitions = None

        return num_replicas, num_partitions

    @classmethod
    def from_response_json(cls, json_dict):
        json_decamelized = humps.decamelize(json_dict)
        return cls.from_json(json_decamelized)

    @classmethod
    def from_json(cls, json_decamelized):
        return KafkaTopic(**cls.extract_fields_from_json(json_decamelized))

    @classmethod
    def extract_fields_from_json(cls, json_decamelized):
        kwargs = {}
        kwargs["name"] = json_decamelized.pop("name")  # required
        kwargs["num_replicas"] = util.extract_field_from_json(
            json_decamelized, ["num_of_replicas", "num_replicas"]
        )
        kwargs["num_partitions"] = util.extract_field_from_json(
            json_decamelized, ["num_of_partitions", "num_partitions"]
        )
        return kwargs

    def update_from_response_json(self, json_dict):
        json_decamelized = humps.decamelize(json_dict)
        self.__init__(**self.extract_fields_from_json(json_decamelized))
        return self

    def json(self):
        return json.dumps(self, cls=util.MLEncoder)

    def to_dict(self):
        return {
            "kafkaTopicDTO": {
                "name": self._name,
                "numOfReplicas": self._num_replicas,
                "numOfPartitions": self._num_partitions,
            }
        }

    @property
    def name(self):
        """Name of the Kafka topic."""
        return self._name

    @name.setter
    def name(self, name: str):
        self._name = name

    @property
    def num_replicas(self):
        """Number of replicas of the Kafka topic."""
        return self._num_replicas

    @num_replicas.setter
    def num_replicas(self, num_replicas: int):
        self._num_replicas = num_replicas

    @property
    def num_partitions(self):
        """Number of partitions of the Kafka topic."""
        return self._num_partitions

    @num_partitions.setter
    def topic_num_partitions(self, num_partitions: int):
        self._num_partitions = num_partitions

    def __repr__(self):
        return f"KafkaTopic({self._name!r})"



#
#   Copyright 2022 Logical Clocks AB
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

import json
from typing import Optional, Union

import humps
from hsml import client, constants, deployment, util
from hsml.constants import (
    ARTIFACT_VERSION,
    INFERENCE_ENDPOINTS,
    MODEL,
    PREDICTOR,
    RESOURCES,
    Default,
)
from hsml.deployable_component import DeployableComponent
from hsml.inference_batcher import InferenceBatcher
from hsml.inference_logger import InferenceLogger
from hsml.predictor_state import PredictorState
from hsml.resources import PredictorResources
from hsml.transformer import Transformer


class Predictor(DeployableComponent):
    """Metadata object representing a predictor in Model Serving."""

    def __init__(
        self,
        name: str,
        model_name: str,
        model_path: str,
        model_version: int,
        model_framework: str,  # MODEL.FRAMEWORK
        artifact_version: Union[int, str],
        model_server: str,
        serving_tool: Optional[str] = None,
        script_file: Optional[str] = None,
        resources: Optional[Union[PredictorResources, dict, Default]] = None,  # base
        inference_logger: Optional[
            Union[InferenceLogger, dict, Default]
        ] = None,  # base
        inference_batcher: Optional[
            Union[InferenceBatcher, dict, Default]
        ] = None,  # base
        transformer: Optional[Union[Transformer, dict, Default]] = None,
        id: Optional[int] = None,
        description: Optional[str] = None,
        created_at: Optional[str] = None,
        creator: Optional[str] = None,
        api_protocol: Optional[str] = INFERENCE_ENDPOINTS.API_PROTOCOL_REST,
        environment: Optional[str] = None,
        project_namespace: str = None,
        **kwargs,
    ):
        serving_tool = (
            self._validate_serving_tool(serving_tool)
            or self._get_default_serving_tool()
        )
        resources = self._validate_resources(
            util.get_obj_from_json(resources, PredictorResources), serving_tool
        ) or self._get_default_resources(serving_tool)

        super().__init__(
            script_file,
            resources,
            inference_batcher,
        )

        self._name = name
        self._model_name = model_name
        self._model_path = model_path
        self._model_version = model_version
        self._model_framework = model_framework
        self._artifact_version = artifact_version
        self._serving_tool = serving_tool
        self._model_server = model_server
        self._id = id
        self._description = description
        self._created_at = created_at
        self._creator = creator

        self._inference_logger = util.get_obj_from_json(
            inference_logger, InferenceLogger
        )
        self._transformer = util.get_obj_from_json(transformer, Transformer)
        self._validate_script_file(self._model_framework, self._script_file)
        self._api_protocol = api_protocol
        self._environment = environment
        self._project_namespace = project_namespace

    def deploy(self):
        """Create a deployment for this predictor and persists it in the Model Serving.

        !!! example
            ```python

            import hopsworks

            project = hopsworks.login()

            # get Hopsworks Model Registry handle
            mr = project.get_model_registry()

            # retrieve the trained model you want to deploy
            my_model = mr.get_model("my_model", version=1)

            # get Hopsworks Model Serving handle
            ms = project.get_model_serving()

            my_predictor = ms.create_predictor(my_model)
            my_deployment = my_predictor.deploy()

            print(my_deployment.get_state())
            ```

        # Returns
            `Deployment`. The deployment metadata object of a new or existing deployment.
        """

        _deployment = deployment.Deployment(
            predictor=self, name=self._name, description=self._description
        )
        _deployment.save()

        return _deployment

    def describe(self):
        """Print a description of the predictor"""
        util.pretty_print(self)

    def _set_state(self, state: PredictorState):
        """Set the state of the predictor"""
        self._state = state

    @classmethod
    def _validate_serving_tool(cls, serving_tool):
        if serving_tool is not None:
            if client.is_saas_connection():
                # only kserve supported in saasy hopsworks
                if serving_tool != PREDICTOR.SERVING_TOOL_KSERVE:
                    raise ValueError(
                        "KServe deployments are the only supported in Serverless Hopsworks"
                    )
                return serving_tool
            # if not saas, check valid serving_tool
            serving_tools = list(util.get_members(PREDICTOR, prefix="SERVING_TOOL"))
            if serving_tool not in serving_tools:
                raise ValueError(
                    "Serving tool '{}' is not valid. Possible values are '{}'".format(
                        serving_tool, ", ".join(serving_tools)
                    )
                )
        return serving_tool

    @classmethod
    def _validate_script_file(cls, model_framework, script_file):
        if script_file is None and (
            model_framework == MODEL.FRAMEWORK_PYTHON
            or model_framework == MODEL.FRAMEWORK_LLM
        ):
            raise ValueError(
                "Predictor scripts are required in deployments for custom Python models and LLMs."
            )

    @classmethod
    def _infer_model_server(cls, model_framework):
        if model_framework == MODEL.FRAMEWORK_TENSORFLOW:
            return PREDICTOR.MODEL_SERVER_TF_SERVING
        elif model_framework == MODEL.FRAMEWORK_LLM:
            return PREDICTOR.MODEL_SERVER_VLLM
        else:
            return PREDICTOR.MODEL_SERVER_PYTHON

    @classmethod
    def _get_default_serving_tool(cls):
        # set kserve as default if it is available
        return (
            PREDICTOR.SERVING_TOOL_KSERVE
            if client.is_kserve_installed()
            else PREDICTOR.SERVING_TOOL_DEFAULT
        )

    @classmethod
    def _validate_resources(cls, resources, serving_tool):
        if resources is not None:
            # ensure scale-to-zero for kserve deployments when required
            if (
                serving_tool == PREDICTOR.SERVING_TOOL_KSERVE
                and resources.num_instances != 0
                and client.is_scale_to_zero_required()
            ):
                raise ValueError(
                    "Scale-to-zero is required for KServe deployments in this cluster. Please, set the number of instances to 0."
                )
        return resources

    @classmethod
    def _get_default_resources(cls, serving_tool):
        num_instances = (
            0  # enable scale-to-zero by default if required
            if serving_tool == PREDICTOR.SERVING_TOOL_KSERVE
            and client.is_scale_to_zero_required()
            else RESOURCES.MIN_NUM_INSTANCES
        )
        return PredictorResources(num_instances)

    @classmethod
    def for_model(cls, model, **kwargs):
        kwargs["model_name"] = model.name
        kwargs["model_path"] = model.model_path
        kwargs["model_version"] = model.version

        # get predictor for specific model, includes model type-related validations
        return util.get_predictor_for_model(model=model, **kwargs)

    @classmethod
    def from_response_json(cls, json_dict):
        json_decamelized = humps.decamelize(json_dict)
        if isinstance(json_decamelized, list):
            if len(json_decamelized) == 0:
                return []
            return [cls.from_json(predictor) for predictor in json_decamelized]
        else:
            if "count" in json_decamelized:
                if json_decamelized["count"] == 0:
                    return []
                return [
                    cls.from_json(predictor) for predictor in json_decamelized["items"]
                ]
            return cls.from_json(json_decamelized)

    @classmethod
    def from_json(cls, json_decamelized):
        predictor = Predictor(**cls.extract_fields_from_json(json_decamelized))
        predictor._set_state(PredictorState.from_response_json(json_decamelized))
        return predictor

    @classmethod
    def extract_fields_from_json(cls, json_decamelized):
        kwargs = {}
        kwargs["name"] = json_decamelized.pop("name")
        kwargs["description"] = util.extract_field_from_json(
            json_decamelized, "description"
        )
        kwargs["model_name"] = util.extract_field_from_json(
            json_decamelized, "model_name", default=kwargs["name"]
        )
        kwargs["model_path"] = json_decamelized.pop("model_path")
        kwargs["model_version"] = json_decamelized.pop("model_version")
        kwargs["model_framework"] = (
            json_decamelized.pop("model_framework")
            if "model_framework" in json_decamelized
            else MODEL.FRAMEWORK_SKLEARN  # backward compatibility
        )
        kwargs["artifact_version"] = util.extract_field_from_json(
            json_decamelized, "artifact_version"
        )
        kwargs["model_server"] = json_decamelized.pop("model_server")
        kwargs["serving_tool"] = json_decamelized.pop("serving_tool")
        kwargs["script_file"] = util.extract_field_from_json(
            json_decamelized, "predictor"
        )
        kwargs["resources"] = PredictorResources.from_json(json_decamelized)
        kwargs["inference_logger"] = InferenceLogger.from_json(json_decamelized)
        kwargs["inference_batcher"] = InferenceBatcher.from_json(json_decamelized)
        kwargs["transformer"] = Transformer.from_json(json_decamelized)
        kwargs["id"] = json_decamelized.pop("id")
        kwargs["created_at"] = json_decamelized.pop("created")
        kwargs["creator"] = json_decamelized.pop("creator")
        kwargs["api_protocol"] = json_decamelized.pop("api_protocol")
        if "environment_dto" in json_decamelized:
            environment = json_decamelized.pop("environment_dto")
            kwargs["environment"] = environment["name"]
        kwargs["project_namespace"] = json_decamelized.pop("project_namespace")
        return kwargs

    def update_from_response_json(self, json_dict):
        json_decamelized = humps.decamelize(json_dict)
        self.__init__(**self.extract_fields_from_json(json_decamelized))
        self._set_state(PredictorState.from_response_json(json_decamelized))
        return self

    def json(self):
        return json.dumps(self, cls=util.MLEncoder)

    def to_dict(self):
        json = {
            "id": self._id,
            "name": self._name,
            "description": self._description,
            "modelName": self._model_name,
            "modelPath": self._model_path,
            "modelVersion": self._model_version,
            "modelFramework": self._model_framework,
            "artifactVersion": self._artifact_version,
            "created": self._created_at,
            "creator": self._creator,
            "modelServer": self._model_server,
            "servingTool": self._serving_tool,
            "predictor": self._script_file,
            "apiProtocol": self._api_protocol,
            "projectNamespace": self._project_namespace,
        }
        if self.environment is not None:
            json = {**json, **{"environmentDTO": {"name": self._environment}}}
        if self._resources is not None:
            json = {**json, **self._resources.to_dict()}
        if self._inference_logger is not None:
            json = {**json, **self._inference_logger.to_dict()}
        if self._inference_batcher is not None:
            json = {**json, **self._inference_batcher.to_dict()}
        if self._transformer is not None:
            json = {**json, **self._transformer.to_dict()}
        return json

    @property
    def id(self):
        """Id of the predictor."""
        return self._id

    @property
    def name(self):
        """Name of the predictor."""
        return self._name

    @name.setter
    def name(self, name: str):
        self._name = name

    @property
    def description(self):
        """Description of the predictor."""
        return self._description

    @description.setter
    def description(self, description: str):
        self._description = description

    @property
    def model_name(self):
        """Name of the model deployed by the predictor."""
        return self._model_name

    @model_name.setter
    def model_name(self, model_name: str):
        self._model_name = model_name

    @property
    def model_path(self):
        """Model path deployed by the predictor."""
        return self._model_path

    @model_path.setter
    def model_path(self, model_path: str):
        self._model_path = model_path

    @property
    def model_version(self):
        """Model version deployed by the predictor."""
        return self._model_version

    @model_version.setter
    def model_version(self, model_version: int):
        self._model_version = model_version

    @property
    def model_framework(self):
        """Model framework of the model to be deployed by the predictor."""
        return self._model_framework

    @model_framework.setter
    def model_framework(self, model_framework: str):
        self._model_framework = model_framework
        self._model_server = self._infer_model_server(model_framework)

    @property
    def artifact_version(self):
        """Artifact version deployed by the predictor."""
        return self._artifact_version

    @artifact_version.setter
    def artifact_version(self, artifact_version: Union[int, str]):
        self._artifact_version = artifact_version

    @property
    def artifact_files_path(self):
        return "{}/{}/{}/{}".format(
            self._model_path,
            str(self._model_version),
            constants.MODEL_SERVING.ARTIFACTS_DIR_NAME,
            str(self._artifact_version),
        )

    @property
    def artifact_path(self):
        """Path of the model artifact deployed by the predictor. Resolves to /Projects/{project_name}/Models/{name}/{version}/Artifacts/{artifact_version}/{name}_{version}_{artifact_version}.zip"""
        # TODO: Deprecated
        artifact_name = "{}_{}_{}.zip".format(
            self._model_name, str(self._model_version), str(self._artifact_version)
        )
        return "{}/{}/Artifacts/{}/{}".format(
            self._model_path,
            str(self._model_version),
            str(self._artifact_version),
            artifact_name,
        )

    @property
    def model_server(self):
        """Model server used by the predictor."""
        return self._model_server

    @property
    def serving_tool(self):
        """Serving tool used to run the model server."""
        return self._serving_tool

    @serving_tool.setter
    def serving_tool(self, serving_tool: str):
        self._serving_tool = serving_tool

    @property
    def script_file(self):
        """Script file used to load and run the model."""
        return self._script_file

    @script_file.setter
    def script_file(self, script_file: str):
        self._script_file = script_file
        self._artifact_version = ARTIFACT_VERSION.CREATE

    @property
    def inference_logger(self):
        """Configuration of the inference logger attached to this predictor."""
        return self._inference_logger

    @inference_logger.setter
    def inference_logger(self, inference_logger: InferenceLogger):
        self._inference_logger = inference_logger

    @property
    def transformer(self):
        """Transformer configuration attached to the predictor."""
        return self._transformer

    @transformer.setter
    def transformer(self, transformer: Transformer):
        self._transformer = transformer

    @property
    def created_at(self):
        """Created at date of the predictor."""
        return self._created_at

    @property
    def creator(self):
        """Creator of the predictor."""
        return self._creator

    @property
    def requested_instances(self):
        """Total number of requested instances in the predictor."""
        num_instances = self._resources.num_instances
        if self._transformer is not None:
            num_instances += self._transformer.resources.num_instances
        return num_instances

    @property
    def api_protocol(self):
        """API protocol enabled in the predictor (e.g., HTTP or GRPC)."""
        return self._api_protocol

    @api_protocol.setter
    def api_protocol(self, api_protocol):
        self._api_protocol = api_protocol

    @property
    def environment(self):
        """Name of the inference environment"""
        return self._environment

    @environment.setter
    def environment(self, environment):
        self._environment = environment

    @property
    def project_namespace(self):
        """Kubernetes project namespace"""
        return self._project_namespace

    @project_namespace.setter
    def project_namespace(self, project_namespace):
        self._project_namespace = project_namespace

    def __repr__(self):
        desc = (
            f", description: {self._description!r}"
            if self._description is not None
            else ""
        )
        return f"Predictor(name: {self._name!r}" + desc + ")"



# Copyright 2022 The KServe Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Generated by the gRPC Python protocol compiler plugin. DO NOT EDIT!
"""Client and server classes corresponding to protobuf-defined services."""

import hopsworks_common.client.istio.grpc.inference_client as inference_client
import hopsworks_common.client.istio.grpc.proto.grpc_predict_v2_pb2 as grpc__predict__v2__pb2


class GRPCInferenceServiceStub(object):
    """Inference Server GRPC endpoints."""

    def __init__(self, channel):
        """Constructor.

        Args:
            channel: A grpc.Channel.
        """
        self.ServerLive = channel.unary_unary(
            "/inference.GRPCInferenceService/ServerLive",
            request_serializer=grpc__predict__v2__pb2.ServerLiveRequest.SerializeToString,
            response_deserializer=grpc__predict__v2__pb2.ServerLiveResponse.FromString,
        )
        self.ServerReady = channel.unary_unary(
            "/inference.GRPCInferenceService/ServerReady",
            request_serializer=grpc__predict__v2__pb2.ServerReadyRequest.SerializeToString,
            response_deserializer=grpc__predict__v2__pb2.ServerReadyResponse.FromString,
        )
        self.ModelReady = channel.unary_unary(
            "/inference.GRPCInferenceService/ModelReady",
            request_serializer=grpc__predict__v2__pb2.ModelReadyRequest.SerializeToString,
            response_deserializer=grpc__predict__v2__pb2.ModelReadyResponse.FromString,
        )
        self.ServerMetadata = channel.unary_unary(
            "/inference.GRPCInferenceService/ServerMetadata",
            request_serializer=grpc__predict__v2__pb2.ServerMetadataRequest.SerializeToString,
            response_deserializer=grpc__predict__v2__pb2.ServerMetadataResponse.FromString,
        )
        self.ModelMetadata = channel.unary_unary(
            "/inference.GRPCInferenceService/ModelMetadata",
            request_serializer=grpc__predict__v2__pb2.ModelMetadataRequest.SerializeToString,
            response_deserializer=grpc__predict__v2__pb2.ModelMetadataResponse.FromString,
        )
        self.ModelInfer = channel.unary_unary(
            "/inference.GRPCInferenceService/ModelInfer",
            request_serializer=grpc__predict__v2__pb2.ModelInferRequest.SerializeToString,
            response_deserializer=grpc__predict__v2__pb2.ModelInferResponse.FromString,
        )
        self.RepositoryModelLoad = channel.unary_unary(
            "/inference.GRPCInferenceService/RepositoryModelLoad",
            request_serializer=grpc__predict__v2__pb2.RepositoryModelLoadRequest.SerializeToString,
            response_deserializer=grpc__predict__v2__pb2.RepositoryModelLoadResponse.FromString,
        )
        self.RepositoryModelUnload = channel.unary_unary(
            "/inference.GRPCInferenceService/RepositoryModelUnload",
            request_serializer=grpc__predict__v2__pb2.RepositoryModelUnloadRequest.SerializeToString,
            response_deserializer=grpc__predict__v2__pb2.RepositoryModelUnloadResponse.FromString,
        )


class GRPCInferenceServiceServicer(object):
    """Inference Server GRPC endpoints."""

    def ServerLive(self, request, context):
        """The ServerLive API indicates if the inference server is able to receive
        and respond to metadata and inference requests.
        """
        context.set_code(inference_client.StatusCode.UNIMPLEMENTED)
        context.set_details("Method not implemented!")
        raise NotImplementedError("Method not implemented!")

    def ServerReady(self, request, context):
        """The ServerReady API indicates if the server is ready for inferencing."""
        context.set_code(inference_client.StatusCode.UNIMPLEMENTED)
        context.set_details("Method not implemented!")
        raise NotImplementedError("Method not implemented!")

    def ModelReady(self, request, context):
        """The ModelReady API indicates if a specific model is ready for inferencing."""
        context.set_code(inference_client.StatusCode.UNIMPLEMENTED)
        context.set_details("Method not implemented!")
        raise NotImplementedError("Method not implemented!")

    def ServerMetadata(self, request, context):
        """The ServerMetadata API provides information about the server. Errors are
        indicated by the google.rpc.Status returned for the request. The OK code
        indicates success and other codes indicate failure.
        """
        context.set_code(inference_client.StatusCode.UNIMPLEMENTED)
        context.set_details("Method not implemented!")
        raise NotImplementedError("Method not implemented!")

    def ModelMetadata(self, request, context):
        """The per-model metadata API provides information about a model. Errors are
        indicated by the google.rpc.Status returned for the request. The OK code
        indicates success and other codes indicate failure.
        """
        context.set_code(inference_client.StatusCode.UNIMPLEMENTED)
        context.set_details("Method not implemented!")
        raise NotImplementedError("Method not implemented!")

    def ModelInfer(self, request, context):
        """The ModelInfer API performs inference using the specified model. Errors are
        indicated by the google.rpc.Status returned for the request. The OK code
        indicates success and other codes indicate failure.
        """
        context.set_code(inference_client.StatusCode.UNIMPLEMENTED)
        context.set_details("Method not implemented!")
        raise NotImplementedError("Method not implemented!")

    def RepositoryModelLoad(self, request, context):
        """Load or reload a model from a repository."""
        context.set_code(inference_client.StatusCode.UNIMPLEMENTED)
        context.set_details("Method not implemented!")
        raise NotImplementedError("Method not implemented!")

    def RepositoryModelUnload(self, request, context):
        """Unload a model."""
        context.set_code(inference_client.StatusCode.UNIMPLEMENTED)
        context.set_details("Method not implemented!")
        raise NotImplementedError("Method not implemented!")


def add_GRPCInferenceServiceServicer_to_server(servicer, server):
    rpc_method_handlers = {
        "ServerLive": inference_client.unary_unary_rpc_method_handler(
            servicer.ServerLive,
            request_deserializer=grpc__predict__v2__pb2.ServerLiveRequest.FromString,
            response_serializer=grpc__predict__v2__pb2.ServerLiveResponse.SerializeToString,
        ),
        "ServerReady": inference_client.unary_unary_rpc_method_handler(
            servicer.ServerReady,
            request_deserializer=grpc__predict__v2__pb2.ServerReadyRequest.FromString,
            response_serializer=grpc__predict__v2__pb2.ServerReadyResponse.SerializeToString,
        ),
        "ModelReady": inference_client.unary_unary_rpc_method_handler(
            servicer.ModelReady,
            request_deserializer=grpc__predict__v2__pb2.ModelReadyRequest.FromString,
            response_serializer=grpc__predict__v2__pb2.ModelReadyResponse.SerializeToString,
        ),
        "ServerMetadata": inference_client.unary_unary_rpc_method_handler(
            servicer.ServerMetadata,
            request_deserializer=grpc__predict__v2__pb2.ServerMetadataRequest.FromString,
            response_serializer=grpc__predict__v2__pb2.ServerMetadataResponse.SerializeToString,
        ),
        "ModelMetadata": inference_client.unary_unary_rpc_method_handler(
            servicer.ModelMetadata,
            request_deserializer=grpc__predict__v2__pb2.ModelMetadataRequest.FromString,
            response_serializer=grpc__predict__v2__pb2.ModelMetadataResponse.SerializeToString,
        ),
        "ModelInfer": inference_client.unary_unary_rpc_method_handler(
            servicer.ModelInfer,
            request_deserializer=grpc__predict__v2__pb2.ModelInferRequest.FromString,
            response_serializer=grpc__predict__v2__pb2.ModelInferResponse.SerializeToString,
        ),
        "RepositoryModelLoad": inference_client.unary_unary_rpc_method_handler(
            servicer.RepositoryModelLoad,
            request_deserializer=grpc__predict__v2__pb2.RepositoryModelLoadRequest.FromString,
            response_serializer=grpc__predict__v2__pb2.RepositoryModelLoadResponse.SerializeToString,
        ),
        "RepositoryModelUnload": inference_client.unary_unary_rpc_method_handler(
            servicer.RepositoryModelUnload,
            request_deserializer=grpc__predict__v2__pb2.RepositoryModelUnloadRequest.FromString,
            response_serializer=grpc__predict__v2__pb2.RepositoryModelUnloadResponse.SerializeToString,
        ),
    }
    generic_handler = inference_client.method_handlers_generic_handler(
        "inference.GRPCInferenceService", rpc_method_handlers
    )
    server.add_generic_rpc_handlers((generic_handler,))


# This class is part of an EXPERIMENTAL API.
class GRPCInferenceService(object):
    """Inference Server GRPC endpoints."""

    @staticmethod
    def ServerLive(
        request,
        target,
        options=(),
        channel_credentials=None,
        call_credentials=None,
        insecure=False,
        compression=None,
        wait_for_ready=None,
        timeout=None,
        metadata=None,
    ):
        return inference_client.experimental.unary_unary(
            request,
            target,
            "/inference.GRPCInferenceService/ServerLive",
            grpc__predict__v2__pb2.ServerLiveRequest.SerializeToString,
            grpc__predict__v2__pb2.ServerLiveResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
        )

    @staticmethod
    def ServerReady(
        request,
        target,
        options=(),
        channel_credentials=None,
        call_credentials=None,
        insecure=False,
        compression=None,
        wait_for_ready=None,
        timeout=None,
        metadata=None,
    ):
        return inference_client.experimental.unary_unary(
            request,
            target,
            "/inference.GRPCInferenceService/ServerReady",
            grpc__predict__v2__pb2.ServerReadyRequest.SerializeToString,
            grpc__predict__v2__pb2.ServerReadyResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
        )

    @staticmethod
    def ModelReady(
        request,
        target,
        options=(),
        channel_credentials=None,
        call_credentials=None,
        insecure=False,
        compression=None,
        wait_for_ready=None,
        timeout=None,
        metadata=None,
    ):
        return inference_client.experimental.unary_unary(
            request,
            target,
            "/inference.GRPCInferenceService/ModelReady",
            grpc__predict__v2__pb2.ModelReadyRequest.SerializeToString,
            grpc__predict__v2__pb2.ModelReadyResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
        )

    @staticmethod
    def ServerMetadata(
        request,
        target,
        options=(),
        channel_credentials=None,
        call_credentials=None,
        insecure=False,
        compression=None,
        wait_for_ready=None,
        timeout=None,
        metadata=None,
    ):
        return inference_client.experimental.unary_unary(
            request,
            target,
            "/inference.GRPCInferenceService/ServerMetadata",
            grpc__predict__v2__pb2.ServerMetadataRequest.SerializeToString,
            grpc__predict__v2__pb2.ServerMetadataResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
        )

    @staticmethod
    def ModelMetadata(
        request,
        target,
        options=(),
        channel_credentials=None,
        call_credentials=None,
        insecure=False,
        compression=None,
        wait_for_ready=None,
        timeout=None,
        metadata=None,
    ):
        return inference_client.experimental.unary_unary(
            request,
            target,
            "/inference.GRPCInferenceService/ModelMetadata",
            grpc__predict__v2__pb2.ModelMetadataRequest.SerializeToString,
            grpc__predict__v2__pb2.ModelMetadataResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
        )

    @staticmethod
    def ModelInfer(
        request,
        target,
        options=(),
        channel_credentials=None,
        call_credentials=None,
        insecure=False,
        compression=None,
        wait_for_ready=None,
        timeout=None,
        metadata=None,
    ):
        return inference_client.experimental.unary_unary(
            request,
            target,
            "/inference.GRPCInferenceService/ModelInfer",
            grpc__predict__v2__pb2.ModelInferRequest.SerializeToString,
            grpc__predict__v2__pb2.ModelInferResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
        )

    @staticmethod
    def RepositoryModelLoad(
        request,
        target,
        options=(),
        channel_credentials=None,
        call_credentials=None,
        insecure=False,
        compression=None,
        wait_for_ready=None,
        timeout=None,
        metadata=None,
    ):
        return inference_client.experimental.unary_unary(
            request,
            target,
            "/inference.GRPCInferenceService/RepositoryModelLoad",
            grpc__predict__v2__pb2.RepositoryModelLoadRequest.SerializeToString,
            grpc__predict__v2__pb2.RepositoryModelLoadResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
        )

    @staticmethod
    def RepositoryModelUnload(
        request,
        target,
        options=(),
        channel_credentials=None,
        call_credentials=None,
        insecure=False,
        compression=None,
        wait_for_ready=None,
        timeout=None,
        metadata=None,
    ):
        return inference_client.experimental.unary_unary(
            request,
            target,
            "/inference.GRPCInferenceService/RepositoryModelUnload",
            grpc__predict__v2__pb2.RepositoryModelUnloadRequest.SerializeToString,
            grpc__predict__v2__pb2.RepositoryModelUnloadResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
        )



#
#   Copyright 2024 Hopsworks AB
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#

import copy

import humps
import pytest
from hsml import inference_logger, kafka_topic
from hsml.constants import DEFAULT, INFERENCE_LOGGER


class TestInferenceLogger:
    # from response json

    def test_from_response_json_with_mode_only(self, mocker, backend_fixtures):
        # Arrange
        json = backend_fixtures["inference_logger"]["get_mode_all"]["response"]
        json_camelized = humps.camelize(json)  # as returned by the backend
        mock_il_from_json = mocker.patch(
            "hsml.inference_logger.InferenceLogger.from_json"
        )

        # Act
        _ = inference_logger.InferenceLogger.from_response_json(json_camelized)

        # Assert
        mock_il_from_json.assert_called_once_with(json)

    def test_from_response_json_with_mode_and_kafka_topic(
        self, mocker, backend_fixtures
    ):
        # Arrange
        json = backend_fixtures["inference_logger"]["get_mode_all_with_kafka_topic"][
            "response"
        ]
        json_camelized = humps.camelize(json)  # as returned by the backend
        mock_il_from_json = mocker.patch(
            "hsml.inference_logger.InferenceLogger.from_json"
        )

        # Act
        _ = inference_logger.InferenceLogger.from_response_json(json_camelized)

        # Assert
        mock_il_from_json.assert_called_once_with(json)

    # from json

    def test_from_json_with_mode_all(self, mocker, backend_fixtures):
        # Arrange
        json = backend_fixtures["inference_logger"]["get_mode_all"]["response"]
        mock_il_extract_fields = mocker.patch(
            "hsml.inference_logger.InferenceLogger.extract_fields_from_json",
            return_value=json,
        )
        mock_il_init = mocker.patch(
            "hsml.inference_logger.InferenceLogger.__init__", return_value=None
        )

        # Act
        _ = inference_logger.InferenceLogger.from_json(json)

        # Assert
        mock_il_extract_fields.assert_called_once_with(json)
        mock_il_init.assert_called_once_with(**json)

    def test_from_json_with_mode_all_and_kafka_topic(self, mocker, backend_fixtures):
        # Arrange
        json = backend_fixtures["inference_logger"]["get_mode_all_with_kafka_topic"][
            "response"
        ]
        mock_il_extract_fields = mocker.patch(
            "hsml.inference_logger.InferenceLogger.extract_fields_from_json",
            return_value=json,
        )
        mock_il_init = mocker.patch(
            "hsml.inference_logger.InferenceLogger.__init__", return_value=None
        )

        # Act
        _ = inference_logger.InferenceLogger.from_json(json)

        # Assert
        mock_il_extract_fields.assert_called_once_with(json)
        mock_il_init.assert_called_once_with(**json)

    # constructor

    def test_constructor_default(self, mocker):
        # Arrange
        mock_il_validate_mode = mocker.patch(
            "hsml.inference_logger.InferenceLogger._validate_mode",
            return_value=INFERENCE_LOGGER.MODE_ALL,
        )
        default_kt = kafka_topic.KafkaTopic()
        mock_util_get_obj_from_json = mocker.patch(
            "hsml.util.get_obj_from_json", return_value=default_kt
        )

        # Act
        il = inference_logger.InferenceLogger()

        # Assert
        assert isinstance(il, inference_logger.InferenceLogger)
        assert il.mode == INFERENCE_LOGGER.MODE_ALL
        assert isinstance(il.kafka_topic, kafka_topic.KafkaTopic)
        assert il.kafka_topic.name == default_kt.name
        assert il.kafka_topic.num_replicas == default_kt.num_replicas
        assert il.kafka_topic.num_partitions == default_kt.num_partitions
        mock_util_get_obj_from_json.assert_called_once_with(
            DEFAULT, kafka_topic.KafkaTopic
        )
        mock_il_validate_mode.assert_called_once_with(
            INFERENCE_LOGGER.MODE_ALL, default_kt
        )

    def test_constructor_mode_all(self, mocker, backend_fixtures):
        # Arrange
        json = backend_fixtures["inference_logger"]["get_mode_all"]["init_args"]
        mock_il_validate_mode = mocker.patch(
            "hsml.inference_logger.InferenceLogger._validate_mode",
            return_value=json["mode"],
        )
        default_kt = kafka_topic.KafkaTopic()
        mock_util_get_obj_from_json = mocker.patch(
            "hsml.util.get_obj_from_json", return_value=default_kt
        )

        # Act
        il = inference_logger.InferenceLogger(**json)

        # Assert
        assert isinstance(il, inference_logger.InferenceLogger)
        assert il.mode == json["mode"]
        assert isinstance(il.kafka_topic, kafka_topic.KafkaTopic)
        assert il.kafka_topic.name == default_kt.name
        assert il.kafka_topic.num_replicas == default_kt.num_replicas
        assert il.kafka_topic.num_partitions == default_kt.num_partitions
        mock_util_get_obj_from_json.assert_called_once_with(
            DEFAULT, kafka_topic.KafkaTopic
        )
        mock_il_validate_mode.assert_called_once_with(json["mode"], default_kt)

    def test_constructor_mode_inputs(self, mocker, backend_fixtures):
        # Arrange
        json = backend_fixtures["inference_logger"]["get_mode_inputs"]["init_args"]
        mock_il_validate_mode = mocker.patch(
            "hsml.inference_logger.InferenceLogger._validate_mode",
            return_value=json["mode"],
        )
        default_kt = kafka_topic.KafkaTopic()
        mock_util_get_obj_from_json = mocker.patch(
            "hsml.util.get_obj_from_json", return_value=default_kt
        )

        # Act
        il = inference_logger.InferenceLogger(**json)

        # Assert
        assert isinstance(il, inference_logger.InferenceLogger)
        assert il.mode == json["mode"]
        assert isinstance(il.kafka_topic, kafka_topic.KafkaTopic)
        assert il.kafka_topic.name == default_kt.name
        assert il.kafka_topic.num_replicas == default_kt.num_replicas
        assert il.kafka_topic.num_partitions == default_kt.num_partitions
        mock_util_get_obj_from_json.assert_called_once_with(
            DEFAULT, kafka_topic.KafkaTopic
        )
        mock_il_validate_mode.assert_called_once_with(json["mode"], default_kt)

    def test_constructor_mode_outputs(self, mocker, backend_fixtures):
        # Arrange
        json = backend_fixtures["inference_logger"]["get_mode_outputs"]["init_args"]
        mock_il_validate_mode = mocker.patch(
            "hsml.inference_logger.InferenceLogger._validate_mode",
            return_value=json["mode"],
        )
        default_kt = kafka_topic.KafkaTopic()
        mock_util_get_obj_from_json = mocker.patch(
            "hsml.util.get_obj_from_json", return_value=default_kt
        )

        # Act
        il = inference_logger.InferenceLogger(**json)

        # Assert
        assert isinstance(il, inference_logger.InferenceLogger)
        assert il.mode == json["mode"]
        assert isinstance(il.kafka_topic, kafka_topic.KafkaTopic)
        assert il.kafka_topic.name == default_kt.name
        assert il.kafka_topic.num_replicas == default_kt.num_replicas
        assert il.kafka_topic.num_partitions == default_kt.num_partitions
        mock_util_get_obj_from_json.assert_called_once_with(
            DEFAULT, kafka_topic.KafkaTopic
        )
        mock_il_validate_mode.assert_called_once_with(json["mode"], default_kt)

    def test_constructor_mode_all_and_kafka_topic(self, mocker, backend_fixtures):
        # Arrange
        json = backend_fixtures["inference_logger"]["get_mode_all_with_kafka_topic"][
            "init_args"
        ]
        json_copy = copy.deepcopy(json)
        mock_il_validate_mode = mocker.patch(
            "hsml.inference_logger.InferenceLogger._validate_mode",
            return_value=json["mode"],
        )
        kt = kafka_topic.KafkaTopic(json["kafka_topic"]["name"])
        mock_util_get_obj_from_json = mocker.patch(
            "hsml.util.get_obj_from_json", return_value=kt
        )

        # Act
        il = inference_logger.InferenceLogger(**json_copy)

        # Assert
        assert isinstance(il, inference_logger.InferenceLogger)
        assert il.mode == json["mode"]
        assert isinstance(il.kafka_topic, kafka_topic.KafkaTopic)
        assert il.kafka_topic.name == kt.name
        assert il.kafka_topic.num_replicas is None
        assert il.kafka_topic.num_partitions is None
        mock_util_get_obj_from_json.assert_called_once_with(
            json["kafka_topic"], kafka_topic.KafkaTopic
        )
        mock_il_validate_mode.assert_called_once_with(json["mode"], kt)

    def test_constructor_mode_inputs_and_kafka_topic(self, mocker, backend_fixtures):
        # Arrange
        json = backend_fixtures["inference_logger"]["get_mode_inputs_with_kafka_topic"][
            "init_args"
        ]
        json_copy = copy.deepcopy(json)
        mock_il_validate_mode = mocker.patch(
            "hsml.inference_logger.InferenceLogger._validate_mode",
            return_value=json["mode"],
        )
        kt = kafka_topic.KafkaTopic(json["kafka_topic"]["name"])
        mock_util_get_obj_from_json = mocker.patch(
            "hsml.util.get_obj_from_json", return_value=kt
        )

        # Act
        il = inference_logger.InferenceLogger(**json_copy)

        # Assert
        assert isinstance(il, inference_logger.InferenceLogger)
        assert il.mode == json["mode"]
        assert isinstance(il.kafka_topic, kafka_topic.KafkaTopic)
        assert il.kafka_topic.name == kt.name
        assert il.kafka_topic.num_replicas is None
        assert il.kafka_topic.num_partitions is None
        mock_util_get_obj_from_json.assert_called_once_with(
            json["kafka_topic"], kafka_topic.KafkaTopic
        )
        mock_il_validate_mode.assert_called_once_with(json["mode"], kt)

    def test_constructor_mode_outputs_and_kafka_topic(self, mocker, backend_fixtures):
        # Arrange
        json = backend_fixtures["inference_logger"]["get_mode_all_with_kafka_topic"][
            "init_args"
        ]
        json_copy = copy.deepcopy(json)
        mock_il_validate_mode = mocker.patch(
            "hsml.inference_logger.InferenceLogger._validate_mode",
            return_value=json["mode"],
        )
        kt = kafka_topic.KafkaTopic(json["kafka_topic"]["name"])
        mock_util_get_obj_from_json = mocker.patch(
            "hsml.util.get_obj_from_json", return_value=kt
        )

        # Act
        il = inference_logger.InferenceLogger(**json_copy)

        # Assert
        assert isinstance(il, inference_logger.InferenceLogger)
        assert il.mode == json["mode"]
        assert isinstance(il.kafka_topic, kafka_topic.KafkaTopic)
        assert il.kafka_topic.name == kt.name
        assert il.kafka_topic.num_replicas is None
        assert il.kafka_topic.num_partitions is None
        mock_util_get_obj_from_json.assert_called_once_with(
            json["kafka_topic"], kafka_topic.KafkaTopic
        )
        mock_il_validate_mode.assert_called_once_with(json["mode"], kt)

    def test_constructor_mode_none_and_kafka_topic(self, mocker, backend_fixtures):
        # Arrange
        json = backend_fixtures["inference_logger"]["get_mode_none_with_kafka_topic"][
            "init_args"
        ]
        json_copy = copy.deepcopy(json)
        mock_il_validate_mode = mocker.patch(
            "hsml.inference_logger.InferenceLogger._validate_mode",
            return_value=json["mode"],
        )
        kt = kafka_topic.KafkaTopic(json["kafka_topic"]["name"])
        mock_util_get_obj_from_json = mocker.patch(
            "hsml.util.get_obj_from_json", return_value=kt
        )

        # Act
        il = inference_logger.InferenceLogger(**json_copy)

        # Assert
        assert isinstance(il, inference_logger.InferenceLogger)
        assert il.mode == json["mode"]
        assert isinstance(il.kafka_topic, kafka_topic.KafkaTopic)
        assert il.kafka_topic.name == kt.name
        assert il.kafka_topic.num_replicas is None
        assert il.kafka_topic.num_partitions is None
        mock_util_get_obj_from_json.assert_called_once_with(
            json["kafka_topic"], kafka_topic.KafkaTopic
        )
        mock_il_validate_mode.assert_called_once_with(json["mode"], kt)

    # validate mode

    def test_validate_mode_none_and_kafka_topic_none(self):
        # Act
        mode = inference_logger.InferenceLogger._validate_mode(None, None)

        # Assert
        assert mode is None

    def test_validate_mode_all_and_kafka_topic_none(self):
        # Act
        mode = inference_logger.InferenceLogger._validate_mode(
            INFERENCE_LOGGER.MODE_ALL, None
        )

        # Assert
        assert mode is None

    def test_validate_mode_invalid_and_kafka_topic_none(self):
        # Act
        with pytest.raises(ValueError) as e_info:
            _ = inference_logger.InferenceLogger._validate_mode("invalid", None)

        # Assert
        assert "is not valid" in str(e_info.value)

    def test_validate_mode_none_and_kafka_topic(self):
        # Act
        mode = inference_logger.InferenceLogger._validate_mode(
            None, kafka_topic.KafkaTopic()
        )

        # Assert
        assert mode == INFERENCE_LOGGER.MODE_NONE

    def test_validate_mode_all_and_kafka_topic(self):
        # Act
        mode = inference_logger.InferenceLogger._validate_mode(
            INFERENCE_LOGGER.MODE_ALL, kafka_topic.KafkaTopic()
        )

        # Assert
        assert mode == INFERENCE_LOGGER.MODE_ALL

    def test_validate_mode_invalid_and_kafka_topic(self):
        # Act
        with pytest.raises(ValueError) as e_info:
            _ = inference_logger.InferenceLogger._validate_mode(
                "invalid", kafka_topic.KafkaTopic()
            )

        # Assert
        assert "is not valid" in str(e_info.value)

    # extract fields from json

    def test_extract_fields_from_json(self, backend_fixtures):
        # Arrange
        json = backend_fixtures["inference_logger"]["get_mode_all_with_kafka_topic"][
            "response"
        ]
        json_copy = copy.deepcopy(json)

        # Act
        kwargs = inference_logger.InferenceLogger.extract_fields_from_json(json_copy)

        # Assert
        assert kwargs["kafka_topic"] == json["kafka_topic_dto"]
        assert kwargs["mode"] == json["inference_logging"]

    def test_extract_fields_from_json_alternative(self, backend_fixtures):
        # Arrange
        json = backend_fixtures["inference_logger"]["get_mode_all_with_kafka_topic"][
            "init_args"
        ]
        json_copy = copy.deepcopy(json)

        # Act
        kwargs = inference_logger.InferenceLogger.extract_fields_from_json(json_copy)

        # Assert
        assert kwargs["kafka_topic"] == json["kafka_topic"]
        assert kwargs["mode"] == json["mode"]



import json
from typing import Any, Dict

import humps
from hsfs import feature_group, util


class FeatureLogging:
    def __init__(
        self,
        id: int,
        transformed_features: "feature_group.FeatureGroup",
        untransformed_features: "feature_group.FeatureGroup",
    ):
        self._id = id
        self._transformed_features = transformed_features
        self._untransformed_features = untransformed_features

    @classmethod
    def from_response_json(cls, json_dict: Dict[str, Any]) -> "FeatureLogging":
        from hsfs.feature_group import FeatureGroup  # avoid circular import

        json_decamelized = humps.decamelize(json_dict)
        transformed_features = json_decamelized.get("transformed_log_fg")
        untransformed_features = json_decamelized.get("untransformed_log_fg")
        if transformed_features:
            transformed_features = FeatureGroup.from_response_json(transformed_features)
        if untransformed_features:
            untransformed_features = FeatureGroup.from_response_json(
                untransformed_features
            )
        return cls(
            json_decamelized.get("id"), transformed_features, untransformed_features
        )

    def update(self, others):
        self._transformed_features = others.transformed_features
        self._untransformed_features = others.untransformed_features
        return self

    @property
    def transformed_features(self) -> "feature_group.FeatureGroup":
        return self._transformed_features

    @property
    def untransformed_features(self) -> "feature_group.FeatureGroup":
        return self._untransformed_features

    def get_feature_group(self, transformed):
        if transformed:
            return self._transformed_features
        else:
            return self._untransformed_features

    @property
    def id(self) -> str:
        return self._id

    def to_dict(self):
        return {
            "id": self._id,
            "transformed_log_fg": self._transformed_features,
            "untransformed_log_fg": self._untransformed_features,
        }

    def json(self) -> Dict[str, Any]:
        return json.dumps(self, cls=util.Encoder)

    def __repr__(self):
        return self.json()



#
#   Copyright 2022 Logical Clocks AB
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#

import json

from hopsworks_common import client, git_provider
from hopsworks_common.client.exceptions import GitException
from hopsworks_common.engine import git_engine


class GitProviderApi:
    def __init__(self):
        self._git_engine = git_engine.GitEngine()

    def _get_providers(self):
        _client = client.get_instance()
        path_params = ["users", "git", "provider"]

        return git_provider.GitProvider.from_response_json(
            _client._send_request("GET", path_params)
        )

    def _get_default_configured_provider(self):
        providers = self._get_providers()
        if providers is None or len(providers) == 0:
            raise GitException("No git provider is configured")
        elif len(providers) == 1:
            return providers[0].git_provider
        else:
            raise GitException(
                "Multiple git providers are configured. Set the provider keyword to specify the provider to use"
            )

    def _get_provider(self, provider: str):
        _client = client.get_instance()
        path_params = ["users", "git", "provider"]

        providers = git_provider.GitProvider.from_response_json(
            _client._send_request("GET", path_params)
        )
        for p in providers:
            if p.git_provider.lower() == provider.lower():
                return p
        raise GitException("No git provider configured for {}".format(provider))

    def _set_provider(self, provider: str, username: str, token: str):
        _client = client.get_instance()
        path_params = ["users", "git", "provider"]

        provider_config = {
            "gitProvider": provider,
            "username": username,
            "token": token,
        }

        headers = {"content-type": "application/json"}
        return git_provider.GitProvider.from_response_json(
            _client._send_request(
                "POST", path_params, headers=headers, data=json.dumps(provider_config)
            )
        )

    def _delete_provider(self, provider: str):
        _client = client.get_instance()
        path_params = ["users", "secrets", "{}_token".format(provider.lower())]
        _client._send_request("DELETE", path_params)
        path_params = ["users", "secrets", "{}_username".format(provider.lower())]
        _client._send_request("DELETE", path_params)



#
#   Copyright 2024 Hopsworks AB
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#

from hopsworks_common.core.constants import (
    HAS_AIOMYSQL,
    HAS_AVRO,
    HAS_CONFLUENT_KAFKA,
    HAS_FAST_AVRO,
    HAS_GREAT_EXPECTATIONS,
    HAS_NUMPY,
    HAS_PANDAS,
    HAS_POLARS,
    HAS_PYARROW,
    HAS_SQLALCHEMY,
    great_expectations_not_installed_message,
    initialise_expectation_suite_for_single_expectation_api_message,
)


__all__ = [
    "HAS_AIOMYSQL",
    "HAS_PYARROW",
    "HAS_AVRO",
    "HAS_CONFLUENT_KAFKA",
    "HAS_FAST_AVRO",
    "HAS_GREAT_EXPECTATIONS",
    "HAS_NUMPY",
    "HAS_PANDAS",
    "HAS_POLARS",
    "HAS_SQLALCHEMY",
    "great_expectations_not_installed_message",
    "initialise_expectation_suite_for_single_expectation_api_message",
]



#
#   Copyright 2021 Logical Clocks AB
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#

import json
import os
import tempfile
import time
import uuid

from hsml import client, constants, util
from hsml.client.exceptions import ModelRegistryException, RestAPIError
from hsml.core import dataset_api, model_api
from hsml.engine import local_engine
from tqdm.auto import tqdm


class ModelEngine:
    def __init__(self):
        self._model_api = model_api.ModelApi()
        self._dataset_api = dataset_api.DatasetApi()

        self._engine = local_engine.LocalEngine()

    def _poll_model_available(self, model_instance, await_registration):
        if await_registration > 0:
            model_registry_id = model_instance.model_registry_id
            sleep_seconds = 5
            for _ in range(int(await_registration / sleep_seconds)):
                try:
                    time.sleep(sleep_seconds)
                    model_meta = self._model_api.get(
                        model_instance.name,
                        model_instance.version,
                        model_registry_id,
                        model_instance.shared_registry_project_name,
                    )
                    if model_meta is not None:
                        return model_meta
                except RestAPIError as e:
                    if e.response.status_code != 404:
                        raise e
            print(
                "Model not available during polling, set a higher value for await_registration to wait longer."
            )

    def _upload_additional_resources(self, model_instance):
        if model_instance._input_example is not None:
            input_example_path = os.path.join(os.getcwd(), "input_example.json")
            input_example = util.input_example_to_json(model_instance._input_example)

            with open(input_example_path, "w+") as out:
                json.dump(input_example, out, cls=util.NumpyEncoder)

            self._engine.upload(input_example_path, model_instance.version_path)
            os.remove(input_example_path)
            model_instance.input_example = None
        if model_instance._model_schema is not None:
            model_schema_path = os.path.join(os.getcwd(), "model_schema.json")
            model_schema = model_instance._model_schema

            with open(model_schema_path, "w+") as out:
                out.write(model_schema.json())

            self._engine.upload(model_schema_path, model_instance.version_path)
            os.remove(model_schema_path)
            model_instance.model_schema = None
        return model_instance

    def _copy_or_move_hopsfs_model_item(
        self, item_attr, to_model_files_path, keep_original_files
    ):
        """Copy or move model item from a hdfs path to the model version folder in the Models dataset. It works with files and folders."""
        path = item_attr["path"]
        to_hdfs_path = os.path.join(to_model_files_path, os.path.basename(path))
        if keep_original_files:
            self._engine.copy(path, to_hdfs_path)
        else:
            self._engine.move(path, to_hdfs_path)

    def _copy_or_move_hopsfs_model(
        self,
        from_hdfs_model_path,
        to_model_files_path,
        keep_original_files,
        update_upload_progress,
    ):
        """Copy or move model files from a hdfs path to the model version folder in the Models dataset."""
        # Strip hdfs prefix
        if from_hdfs_model_path.startswith("hdfs:/"):
            projects_index = from_hdfs_model_path.find("/Projects", 0)
            from_hdfs_model_path = from_hdfs_model_path[projects_index:]

        n_dirs, n_files = 0, 0

        model_path_resp = self._dataset_api.get(from_hdfs_model_path)
        model_path_attr = model_path_resp["attributes"]
        if (
            "datasetType" in model_path_resp
            and model_path_resp["datasetType"] == "DATASET"
        ):  # This is needed to avoid a user exporting for example "Resources" from wiping the dataset
            raise AssertionError(
                "It is disallowed to export a root dataset path."
                " Move the model to a sub-folder and try again."
            )
        elif model_path_attr.get("dir", False):
            # if path is a directory, iterate of the directory content
            for entry in self._dataset_api.list(
                from_hdfs_model_path, sort_by="NAME:desc"
            )["items"]:
                path_attr = entry["attributes"]
                self._copy_or_move_hopsfs_model_item(
                    path_attr, to_model_files_path, keep_original_files
                )
                if path_attr.get("dir", False):
                    n_dirs += 1
                else:
                    n_files += 1
                update_upload_progress(n_dirs=n_dirs, n_files=n_files)
        else:
            # if path is a file, copy/move it
            self._copy_or_move_hopsfs_model_item(
                model_path_attr, to_model_files_path, keep_original_files
            )
            n_files += 1
            update_upload_progress(n_dirs=n_dirs, n_files=n_files)

    def _download_model_from_hopsfs_recursive(
        self,
        from_hdfs_model_path: str,
        to_local_path: str,
        update_download_progress,
        n_dirs,
        n_files,
    ):
        """Download model files from a model path in hdfs, recursively"""

        for entry in self._dataset_api.list(from_hdfs_model_path, sort_by="NAME:desc")[
            "items"
        ]:
            path_attr = entry["attributes"]
            path = path_attr["path"]
            basename = os.path.basename(path)

            if path_attr.get("dir", False):
                # otherwise, make a recursive call for the folder
                if (
                    basename == constants.MODEL_SERVING.ARTIFACTS_DIR_NAME
                ):  # TODO: Not needed anymore
                    continue  # skip Artifacts subfolder
                local_folder_path = os.path.join(to_local_path, basename)
                os.mkdir(local_folder_path)
                n_dirs, n_files = self._download_model_from_hopsfs_recursive(
                    from_hdfs_model_path=path,
                    to_local_path=local_folder_path,
                    update_download_progress=update_download_progress,
                    n_dirs=n_dirs,
                    n_files=n_files,
                )
                n_dirs += 1
                update_download_progress(n_dirs=n_dirs, n_files=n_files)
            else:
                # if it's a file, download it
                local_file_path = os.path.join(to_local_path, basename)
                self._engine.download(path, local_file_path)
                n_files += 1
                update_download_progress(n_dirs=n_dirs, n_files=n_files)

        return n_dirs, n_files

    def _download_model_from_hopsfs(
        self, from_hdfs_model_path: str, to_local_path: str, update_download_progress
    ):
        """Download model files from a model path in hdfs."""

        n_dirs, n_files = self._download_model_from_hopsfs_recursive(
            from_hdfs_model_path=from_hdfs_model_path,
            to_local_path=to_local_path,
            update_download_progress=update_download_progress,
            n_dirs=0,
            n_files=0,
        )
        update_download_progress(n_dirs=n_dirs, n_files=n_files, done=True)

    def _upload_local_model(
        self,
        from_local_model_path,
        to_model_files_path,
        update_upload_progress,
        upload_configuration=None,
    ):
        """Copy or upload model files from a local path to the model files folder in the Models dataset."""
        n_dirs, n_files = 0, 0
        if os.path.isdir(from_local_model_path):
            # if path is a dir, upload files and folders iteratively
            for root, dirs, files in os.walk(from_local_model_path):
                # os.walk(local_model_path), where local_model_path is expected to be an absolute path
                # - root is the absolute path of the directory being walked
                # - dirs is the list of directory names present in the root dir
                # - files is the list of file names present in the root dir
                # we need to replace the local path prefix with the hdfs path prefix (i.e., /srv/hops/....../root with /Projects/.../)
                remote_base_path = root.replace(
                    from_local_model_path, to_model_files_path
                ).replace(os.sep, "/")
                for d_name in dirs:
                    self._engine.mkdir(remote_base_path + "/" + d_name)
                    n_dirs += 1
                    update_upload_progress(n_dirs, n_files)
                for f_name in files:
                    self._engine.upload(
                        root + "/" + f_name,
                        remote_base_path,
                        upload_configuration=upload_configuration,
                    )
                    n_files += 1
                    update_upload_progress(n_dirs, n_files)
        else:
            # if path is a file, upload file
            self._engine.upload(
                from_local_model_path,
                to_model_files_path,
                upload_configuration=upload_configuration,
            )
            n_files += 1
            update_upload_progress(n_dirs, n_files)

    def _save_model_from_local_or_hopsfs_mount(
        self,
        model_instance,
        model_path,
        keep_original_files,
        update_upload_progress,
        upload_configuration=None,
    ):
        """Save model files from a local path. The local path can be on hopsfs mount"""
        # check hopsfs mount
        if model_path.startswith(constants.MODEL_REGISTRY.HOPSFS_MOUNT_PREFIX):
            self._copy_or_move_hopsfs_model(
                from_hdfs_model_path=model_path.replace(
                    constants.MODEL_REGISTRY.HOPSFS_MOUNT_PREFIX, ""
                ),
                to_model_files_path=model_instance.model_files_path,
                keep_original_files=keep_original_files,
                update_upload_progress=update_upload_progress,
            )
        else:
            self._upload_local_model(
                from_local_model_path=model_path,
                to_model_files_path=model_instance.model_files_path,
                update_upload_progress=update_upload_progress,
                upload_configuration=upload_configuration,
            )

    def _set_model_version(
        self, model_instance, dataset_models_root_path, dataset_model_path
    ):
        # Set model version if not defined
        if model_instance._version is None:
            current_highest_version = 0
            for item in self._dataset_api.list(dataset_model_path, sort_by="NAME:desc")[
                "items"
            ]:
                _, file_name = os.path.split(item["attributes"]["path"])
                try:
                    try:
                        current_version = int(file_name)
                    except ValueError:
                        continue
                    if current_version > current_highest_version:
                        current_highest_version = current_version
                except RestAPIError:
                    pass
            model_instance._version = current_highest_version + 1

        elif self._dataset_api.path_exists(
            dataset_models_root_path
            + "/"
            + model_instance._name
            + "/"
            + str(model_instance._version)
        ):
            raise ModelRegistryException(
                "Model with name {} and version {} already exists".format(
                    model_instance._name, model_instance._version
                )
            )
        return model_instance

    def _build_resource_path(self, model_instance, artifact):
        artifact_path = "{}/{}".format(model_instance.version_path, artifact)
        return artifact_path

    def save(
        self,
        model_instance,
        model_path,
        await_registration=480,
        keep_original_files=False,
        upload_configuration=None,
    ):
        _client = client.get_instance()

        is_shared_registry = model_instance.shared_registry_project_name is not None

        if is_shared_registry:
            dataset_models_root_path = "{}::{}".format(
                model_instance.shared_registry_project_name,
                constants.MODEL_SERVING.MODELS_DATASET,
            )
            model_instance._project_name = model_instance.shared_registry_project_name
        else:
            dataset_models_root_path = constants.MODEL_SERVING.MODELS_DATASET
            model_instance._project_name = _client._project_name

        util.validate_metrics(model_instance.training_metrics)

        if not self._dataset_api.path_exists(dataset_models_root_path):
            raise AssertionError(
                "{} dataset does not exist in this project. Please enable the Serving service or create it manually.".format(
                    dataset_models_root_path
                )
            )

        # Create /Models/{model_instance._name} folder
        dataset_model_name_path = dataset_models_root_path + "/" + model_instance._name
        if not self._dataset_api.path_exists(dataset_model_name_path):
            self._engine.mkdir(dataset_model_name_path)

        model_instance = self._set_model_version(
            model_instance, dataset_models_root_path, dataset_model_name_path
        )

        # Attach model summary xattr to /Models/{model_instance._name}/{model_instance._version}
        model_query_params = {}

        if "HOPSWORKS_JOB_NAME" in os.environ:
            model_query_params["jobName"] = os.environ["HOPSWORKS_JOB_NAME"]
        elif "HOPSWORKS_KERNEL_ID" in os.environ:
            model_query_params["kernelId"] = os.environ["HOPSWORKS_KERNEL_ID"]

        pbar = tqdm(
            [
                {"id": 0, "desc": "Creating model folder"},
                {"id": 1, "desc": "Uploading model files"},
                {"id": 2, "desc": "Uploading input_example and model_schema"},
                {"id": 3, "desc": "Registering model"},
                {"id": 4, "desc": "Waiting for model registration"},
                {"id": 5, "desc": "Model export complete"},
            ]
        )

        for step in pbar:
            try:
                pbar.set_description("%s" % step["desc"])
                if step["id"] == 0:
                    # Create folders
                    self._engine.mkdir(model_instance.version_path)
                    self._engine.mkdir(model_instance.model_files_path)
                if step["id"] == 1:

                    def update_upload_progress(n_dirs=0, n_files=0, step=step):
                        pbar.set_description(
                            "%s (%s dirs, %s files)" % (step["desc"], n_dirs, n_files)
                        )

                    update_upload_progress(n_dirs=0, n_files=0)

                    # Upload Model files from local path to /Models/{model_instance._name}/{model_instance._version}/Files
                    # check local absolute
                    if os.path.isabs(model_path) and os.path.exists(model_path):
                        self._save_model_from_local_or_hopsfs_mount(
                            model_instance=model_instance,
                            model_path=model_path,
                            keep_original_files=keep_original_files,
                            update_upload_progress=update_upload_progress,
                            upload_configuration=upload_configuration,
                        )
                    # check local relative
                    elif os.path.exists(
                        os.path.join(os.getcwd(), model_path)
                    ):  # check local relative
                        self._save_model_from_local_or_hopsfs_mount(
                            model_instance=model_instance,
                            model_path=os.path.join(os.getcwd(), model_path),
                            keep_original_files=keep_original_files,
                            update_upload_progress=update_upload_progress,
                            upload_configuration=upload_configuration,
                        )
                    # check project relative
                    elif self._dataset_api.path_exists(
                        model_path
                    ):  # check hdfs relative and absolute
                        self._copy_or_move_hopsfs_model(
                            from_hdfs_model_path=model_path,
                            to_model_files_path=model_instance.model_files_path,
                            keep_original_files=keep_original_files,
                            update_upload_progress=update_upload_progress,
                        )
                    else:
                        raise IOError(
                            "Could not find path {} in the local filesystem or in Hopsworks File System".format(
                                model_path
                            )
                        )
                if step["id"] == 2:
                    model_instance = self._upload_additional_resources(model_instance)
                if step["id"] == 3:
                    model_instance = self._model_api.put(
                        model_instance, model_query_params
                    )
                if step["id"] == 4:
                    model_instance = self._poll_model_available(
                        model_instance, await_registration
                    )
                if step["id"] == 5:
                    pass
            except BaseException as be:
                self._dataset_api.rm(model_instance.version_path)
                raise be

        print("Model created, explore it at " + model_instance.get_url())

        return model_instance

    def download(self, model_instance, local_path=None):
        if local_path is None:
            local_path = os.path.join(
                tempfile.gettempdir(), str(uuid.uuid4()), model_instance._name
            )
            local_path = local_path + "/" + str(model_instance._version)
        os.makedirs(local_path, exist_ok=True)

        def update_download_progress(n_dirs, n_files, done=False):
            print(
                "Downloading model artifact (%s dirs, %s files)... %s"
                % (n_dirs, n_files, "DONE" if done else ""),
                end="\r",
            )

        try:
            from_hdfs_model_path = model_instance.model_files_path
            if from_hdfs_model_path.startswith("hdfs:/"):
                projects_index = from_hdfs_model_path.find("/Projects", 0)
                from_hdfs_model_path = from_hdfs_model_path[projects_index:]

            self._download_model_from_hopsfs(
                from_hdfs_model_path=from_hdfs_model_path,
                to_local_path=local_path,
                update_download_progress=update_download_progress,
            )
        except BaseException as be:
            raise be

        return local_path

    def read_file(self, model_instance, resource):
        hdfs_resource_path = self._build_resource_path(
            model_instance, os.path.basename(resource)
        )
        if self._dataset_api.path_exists(hdfs_resource_path):
            try:
                resource = os.path.basename(resource)
                tmp_dir = tempfile.TemporaryDirectory(dir=os.getcwd())
                local_resource_path = os.path.join(tmp_dir.name, resource)
                self._engine.download(
                    hdfs_resource_path,
                    local_resource_path,
                )
                with open(local_resource_path, "r") as f:
                    return f.read()
            finally:
                if tmp_dir is not None and os.path.exists(tmp_dir.name):
                    tmp_dir.cleanup()

    def read_json(self, model_instance, resource):
        hdfs_resource_path = self._build_resource_path(model_instance, resource)
        if self._dataset_api.path_exists(hdfs_resource_path):
            try:
                tmp_dir = tempfile.TemporaryDirectory(dir=os.getcwd())
                local_resource_path = os.path.join(tmp_dir.name, resource)
                self._engine.download(
                    hdfs_resource_path,
                    local_resource_path,
                )
                with open(local_resource_path, "rb") as f:
                    return json.loads(f.read())
            finally:
                if tmp_dir is not None and os.path.exists(tmp_dir.name):
                    tmp_dir.cleanup()

    def delete(self, model_instance):
        self._engine.delete(model_instance)

    def set_tag(self, model_instance, name, value):
        """Attach a name/value tag to a model."""
        self._model_api.set_tag(model_instance, name, value)

    def delete_tag(self, model_instance, name):
        """Remove a tag from a model."""
        self._model_api.delete_tag(model_instance, name)

    def get_tag(self, model_instance, name):
        """Get tag with a certain name."""
        return self._model_api.get_tags(model_instance, name)[name]

    def get_tags(self, model_instance):
        """Get all tags for a model."""
        return self._model_api.get_tags(model_instance)

    def get_feature_view_provenance(self, model_instance):
        """Get the parent feature view of this model, based on explicit provenance.
        These feature views can be accessible, deleted or inaccessible.
        For deleted and inaccessible feature views, only a minimal information is
        returned.

        # Arguments
            model_instance: Metadata object of model.

        # Returns
            `ProvenanceLinks`:  the feature view used to generate this model
        """
        return self._model_api.get_feature_view_provenance(model_instance)

    def get_training_dataset_provenance(self, model_instance):
        """Get the parent training dataset of this model, based on explicit provenance.
        These training datasets can be accessible, deleted or inaccessible.
        For deleted and inaccessible feature views, only a minimal information is
        returned.

        # Arguments
            model_instance: Metadata object of model.

        # Returns
            `ProvenanceLinks`:  the training dataset used to generate this model
        """
        return self._model_api.get_training_dataset_provenance(model_instance)



#
#   Copyright 2024 Hopsworks AB
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#

import importlib.util


# Avro
HAS_FAST_AVRO: bool = importlib.util.find_spec("fastavro") is not None
HAS_AVRO: bool = importlib.util.find_spec("avro") is not None
avro_not_installed_message = (
    "Avro package not found. "
    "If you want to use avro with Hopsworks you can install the corresponding extra via "
    '`pip install "hopsworks[avro]"`. '
    "You can also install avro directly in your environment with `pip install fastavro` or `pip install avro`. "
    "You will need to restart your kernel if applicable."
)

# Confluent Kafka
HAS_CONFLUENT_KAFKA: bool = importlib.util.find_spec("confluent_kafka") is not None
confluent_kafka_not_installed_message = (
    "Confluent Kafka package not found. "
    "If you want to use Kafka with Hopsworks you can install the corresponding extras via "
    '`pip install "hopsworks[python]"`. '
    "You can also install confluent-kafka directly in your environment with `pip install confluent-kafka`. "
    "You will need to restart your kernel if applicable."
)

# Data Validation / Great Expectations
HAS_GREAT_EXPECTATIONS: bool = (
    importlib.util.find_spec("great_expectations") is not None
)
great_expectations_not_installed_message = (
    "Great Expectations package not found. "
    "If you want to use data validation with Hopsworks you can install the corresponding extras via "
    '`pip install "hopsworks[great_expectations]"`. '
    "You can also install great-expectations directly in your environment with `pip install great-expectations`. "
    "You will need to restart your kernel if applicable."
)
initialise_expectation_suite_for_single_expectation_api_message = "Initialize Expectation Suite by attaching to a Feature Group to enable single expectation API"

# Pyarrow
HAS_PYARROW: bool = importlib.util.find_spec("pyarrow") is not None
pyarrow_not_installed_message = (
    "Pyarrow package not found. "
    "If you want to use Apache Arrow with Hopsworks you can install the corresponding extras via "
    '`pip install "hopsworks[python]"`. '
    "You can also install pyarrow directly in your environment with `pip install pyarrow`. "
    "You will need to restart your kernel if applicable."
)

HAS_PANDAS: bool = importlib.util.find_spec("pandas") is not None

# NumPy
HAS_NUMPY: bool = importlib.util.find_spec("numpy") is not None
numpy_not_installed_message = (
    "Numpy package not found. "
    "If you want to use numpy with Hopsworks you can install the corresponding extra via "
    '`pip install "hopsworks[numpy]"`. '
    "You can also install numpy directly in your environment with `pip install numpy`. "
    "You will need to restart your kernel if applicable."
)

HAS_POLARS: bool = importlib.util.find_spec("polars") is not None
polars_not_installed_message = (
    "Polars package not found. "
    "If you want to use Polars with Hopsworks you can install the corresponding extra via "
    """'`pip install "hopsworks[polars]"`. '"""
    "You can also install polars directly in your environment with `pip install polars`. "
    "You will need to restart your kernel if applicable."
)

# SQL packages
HAS_SQLALCHEMY: bool = importlib.util.find_spec("sqlalchemy") is not None
HAS_AIOMYSQL: bool = importlib.util.find_spec("aiomysql") is not None
